"""Housekeeper fixtures for demultiplex tests."""

from datetime import datetime
from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from tests.store_helpers import StoreHelpers


@pytest.fixture(scope="function")
def illumina_demultiplexed_runs_post_processing_hk_api(
    canonical_flow_cell_ids_and_sample_sheet_paths,
    tmp_fastq_files_for_all_canonical_illumina_demultiplexed_runs: dict[str, list[Path]],
    helpers: StoreHelpers,
    real_housekeeper_api: HousekeeperAPI,
) -> HousekeeperAPI:
    """Return a Housekeeper API instance with Illumina demultiplexed runs."""
    for flow_cell_id in canonical_flow_cell_ids_and_sample_sheet_paths.keys():
        sample_sheet_path: Path = canonical_flow_cell_ids_and_sample_sheet_paths[flow_cell_id]
        run_sample_sheet_bundle: dict = {
            "name": flow_cell_id,
            "created": datetime.now(),
            "version": "1.0",
            "files": [
                {
                    "path": sample_sheet_path.as_posix(),
                    "tags": ["samplesheet", flow_cell_id],
                    "archive": False,
                }
            ],
        }
        helpers.ensure_hk_bundle(store=real_housekeeper_api, bundle_data=run_sample_sheet_bundle)

        sample_ids: list[str] = tmp_fastq_files_for_all_canonical_illumina_demultiplexed_runs[
            flow_cell_id
        ].keys()
        for sample_id in sample_ids:
            tmp_fastq_path: Path = tmp_fastq_files_for_all_canonical_illumina_demultiplexed_runs[
                flow_cell_id
            ][sample_id]
            bundle_data: dict = {
                "name": sample_id,
                "created": datetime.now(),
                "version": "1.0",
                "files": [
                    {
                        "path": tmp_fastq_path.as_posix(),
                        "tags": [SequencingFileTag.FASTQ, flow_cell_id],
                        "archive": False,
                    },
                ],
            }
            helpers.ensure_hk_bundle(store=real_housekeeper_api, bundle_data=bundle_data)
    return real_housekeeper_api



from cg.models.orders.order import OrderIn
from cg.services.orders.store_order_services.store_pacbio_order_service import (
    StorePacBioOrderService,
)
from cg.services.orders.submitters.order_submitter import OrderSubmitter
from cg.services.orders.validate_order_services.validate_pacbio_order import (
    ValidatePacbioOrderService,
)


class PacbioOrderSubmitter(OrderSubmitter):
    """Submitter for Pacbio orders."""

    def __init__(
        self,
        order_validation_service: ValidatePacbioOrderService,
        order_store_service: StorePacBioOrderService,
    ):
        self.order_validation_service = order_validation_service
        self.order_store_service = order_store_service

    def submit_order(self, order_in: OrderIn) -> dict:
        """Submit a fastq order."""
        self.order_validation_service.validate_order(order_in)
        result: dict = self.order_store_service.store_order(order_in)
        return result



import datetime as dt
import getpass
import logging
import os
import re
from pathlib import Path
import paramiko
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.constants import FileExtensions
from cg.constants.constants import SARS_COV_REGEX, DataDelivery
from cg.constants.housekeeper_tags import FohmTag
from cg.exc import CgError
from cg.io.csv import read_csv, write_csv_from_dict
from cg.models.cg_config import CGConfig
from cg.models.email import EmailInfo
from cg.models.fohm.reports import FohmComplementaryReport, FohmPangolinReport
from cg.services.deliver_files.constants import DeliveryDestination, DeliveryStructure
from cg.services.deliver_files.factory import (
    DeliveryServiceFactory,
)
from cg.store.models import Case, Sample
from cg.store.store import Store
from cg.utils.dict import remove_duplicate_dicts
from cg.utils.email import send_mail

LOG = logging.getLogger(__name__)


class FOHMUploadAPI:
    def __init__(
        self,
        config: CGConfig,
        dry_run: bool = False,
        datestr: str | None = None,
    ):
        self.config: CGConfig = config
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.lims_api: LimsAPI = config.lims_api
        self.status_db: Store = config.status_db
        self._dry_run: bool = dry_run
        self._current_datestr: str = datestr or str(dt.date.today())
        self._daily_bundle_path = None
        self._daily_rawdata_path = None
        self._daily_report_path = None
        self._cases_to_aggregate: list[str] = []
        self._daily_reports_list: list[Path] = []
        self._daily_pangolin_list: list[Path] = []
        self._reports_dataframe = None
        self._pangolin_dataframe = None
        self._aggregation_dataframe = None
        self._delivery_factory: DeliveryServiceFactory = config.delivery_service_factory

    @property
    def current_datestr(self) -> str:
        if not self._current_datestr:
            self._current_datestr = str(dt.date.today())
        return self._current_datestr

    @property
    def daily_bundle_path(self) -> Path:
        if not self._daily_bundle_path:
            self._daily_bundle_path: Path = Path(
                Path(self.config.mutant.root).parent, "fohm", self.current_datestr
            )
        return self._daily_bundle_path

    @property
    def daily_rawdata_path(self) -> Path:
        if not self._daily_rawdata_path:
            self._daily_rawdata_path: Path = Path(self.daily_bundle_path, "rawdata")
        return self._daily_rawdata_path

    @property
    def daily_report_path(self) -> Path:
        if not self._daily_report_path:
            self._daily_report_path: Path = Path(self.daily_bundle_path, "reports")
        return self._daily_report_path

    @property
    def daily_reports_list(self) -> list[Path]:
        if not self._daily_reports_list:
            for case_id in self._cases_to_aggregate:
                self._daily_reports_list.append(
                    Path(
                        self.housekeeper_api.get_file_from_latest_version(
                            bundle_name=case_id, tags={FohmTag.COMPLEMENTARY}
                        ).full_path
                    )
                )
        return self._daily_reports_list

    @property
    def daily_pangolin_list(self) -> list[Path]:
        if not self._daily_pangolin_list:
            for case_id in self._cases_to_aggregate:
                self._daily_pangolin_list.append(
                    Path(
                        self.housekeeper_api.get_file_from_latest_version(
                            bundle_name=case_id, tags={FohmTag.PANGOLIN_TYPING}
                        ).full_path
                    )
                )
        return self._daily_pangolin_list

    @property
    def dry_run(self):
        return self._dry_run

    @staticmethod
    def get_reports_contents(file_paths: list[Path]) -> list[dict]:
        """Return a list of dicts with all CSV files reports."""
        csv_reports = [
            read_csv(file_path=file_path, read_to_dict=True, ignore_suffix=True)
            for file_path in file_paths
        ]
        reports: list[dict] = []
        for report in csv_reports:
            reports.extend(iter(report))
        return reports

    @staticmethod
    def validate_fohm_complementary_reports(reports: list[dict]) -> list[FohmComplementaryReport]:
        """Validate FOHM complementary reports."""
        complementary_reports: list[FohmComplementaryReport] = []
        for report in reports:
            complementary_report = FohmComplementaryReport.model_validate(report)
            complementary_reports.append(complementary_report)
        return complementary_reports

    @staticmethod
    def validate_fohm_pangolin_reports(reports: list[dict]) -> list[FohmPangolinReport]:
        """Validate FOHM Pangolin reports."""
        pangolin_reports: list[FohmPangolinReport] = []
        for report in reports:
            pangolin_report = FohmPangolinReport.model_validate(report)
            pangolin_reports.append(pangolin_report)
        return pangolin_reports

    @staticmethod
    def get_sars_cov_complementary_reports(
        reports: list[FohmComplementaryReport],
    ) -> list[FohmComplementaryReport]:
        """Return all "Sars-cov2" reports from reports."""
        return [report for report in reports if re.search(SARS_COV_REGEX, report.sample_number)]

    @staticmethod
    def get_sars_cov_pangolin_reports(
        reports: list[FohmPangolinReport],
    ) -> list[FohmPangolinReport]:
        """Return all "Sars-cov2" reports from reports."""
        return [report for report in reports if re.search(SARS_COV_REGEX, report.taxon)]

    def check_username(self) -> None:
        if self._dry_run:
            return
        if getpass.getuser() == self.config.fohm.valid_uploader:
            return
        raise CgError(
            f"Cannot upload to FOHM as {getpass.getuser()}, please log in as {self.config.fohm.valid_uploader} "
        )

    def set_cases_to_aggregate(self, cases: list) -> None:
        LOG.info(f"Preparing aggregated delivery for {cases}")
        self._cases_to_aggregate = cases

    def create_daily_delivery_folders(self) -> None:
        LOG.info(f"Creating directory: {self.daily_rawdata_path}")
        LOG.info(f"Creating directory: {self.daily_report_path}")
        if self._dry_run:
            return
        self.daily_rawdata_path.mkdir(parents=True, exist_ok=True)
        self.daily_report_path.mkdir(parents=True, exist_ok=True)

    def add_sample_internal_id_to_complementary_reports(
        self, reports: list[FohmComplementaryReport]
    ) -> None:
        """Add sample internal id to complementary reports."""
        for report in reports:
            report.internal_id = self.status_db.get_sample_by_name(
                name=report.sample_number
            ).internal_id

    def add_sample_internal_id_to_pangolin_reports(self, reports: list[FohmPangolinReport]) -> None:
        """Add sample internal id to Pangolin reports."""
        for report in reports:
            report.internal_id = self.status_db.get_sample_by_name(name=report.taxon).internal_id

    def add_region_lab_to_reports(
        self, reports: list[FohmComplementaryReport] | list[FohmPangolinReport]
    ) -> None:
        """Add region laboratory to reports."""
        for report in reports:
            report.region_lab = f"""{self.lims_api.get_sample_attribute(lims_id=report.internal_id, key="region_code").split(' ')[0]}"""
            f"""_{self.lims_api.get_sample_attribute(lims_id=report.internal_id, key="lab_code").split(' ')[0]}"""

    def link_sample_raw_data_files(
        self, reports: list[FohmComplementaryReport] | list[FohmPangolinReport]
    ) -> None:
        """Hardlink samples raw data files to FOHM delivery folder."""
        for report in reports:
            sample: Sample = self.status_db.get_sample_by_internal_id(
                internal_id=report.internal_id
            )
            case: Case = sample.links[0].case
            delivery_service = self._delivery_factory.build_delivery_service(
                case=case,
                delivery_type=DataDelivery.FASTQ_ANALYSIS,
                delivery_destination=DeliveryDestination.FOHM,
                delivery_structure=DeliveryStructure.FLAT,
            )
            delivery_service.deliver_files_for_sample_no_rsync(
                case=case, sample_id=sample.internal_id, delivery_base_path=self.daily_rawdata_path
            )

    def create_pangolin_report(self, reports: list[FohmPangolinReport]) -> None:
        LOG.info("Creating aggregate Pangolin report")
        unique_region_labs: set[str] = {report.region_lab for report in reports}
        LOG.info(f"Regions in batch: {unique_region_labs}")
        for region_lab in unique_region_labs:
            LOG.info(f"Aggregating data for {region_lab}")
            region_lab_reports: list[FohmPangolinReport] = [
                report for report in reports if report.region_lab == region_lab
            ]
            if self._dry_run:
                LOG.info(region_lab_reports)
                continue
            pangolin_report_file = Path(
                self.daily_rawdata_path,
                f"{region_lab}_{self.current_datestr}_pangolin_classification_format4{FileExtensions.TXT}",
            )
            region_lab_reports_raw: list[dict] = [
                report.model_dump() for report in region_lab_reports
            ]
            field_names: list[str] = sorted(region_lab_reports[0].model_dump().keys())
            write_csv_from_dict(
                content=region_lab_reports_raw,
                fieldnames=field_names,
                file_path=pangolin_report_file,
            )
            pangolin_report_file.chmod(0o0777)

    def create_complementary_report(self, reports: list[FohmComplementaryReport]) -> None:
        LOG.info("Creating aggregate 'komplettering' report")
        unique_region_labs: set[str] = {report.region_lab for report in reports}
        LOG.info(f"Regions laboratories  in batch: {unique_region_labs}")
        for region_lab in unique_region_labs:
            LOG.info(f"Aggregating data for {region_lab}")
            region_lab_reports: list[FohmComplementaryReport] = [
                report for report in reports if report.region_lab == region_lab
            ]
            if self._dry_run:
                LOG.info(region_lab_reports)
                continue
            region_lab_reports_raw: list[dict] = [
                report.model_dump(by_alias=True) for report in region_lab_reports
            ]
            field_names: list[str] = sorted(region_lab_reports[0].model_dump(by_alias=True).keys())
            complementary_report_file = Path(
                self.daily_report_path,
                f"{region_lab}_{self.current_datestr}_komplettering{FileExtensions.CSV}",
            )
            write_csv_from_dict(
                content=region_lab_reports_raw,
                fieldnames=field_names,
                file_path=complementary_report_file,
            )

    def send_mail_reports(self) -> None:
        for file in self.daily_report_path.iterdir():
            LOG.info(
                f"Sending {file.name} report file to {self.config.fohm.email_recipient}, dry run: {self._dry_run}"
            )
            if self._dry_run:
                return

            try:
                send_mail(
                    EmailInfo(
                        receiver_email=self.config.fohm.email_recipient,
                        sender_email=self.config.fohm.email_sender,
                        smtp_server=self.config.fohm.email_host,
                        subject=file.name,
                        message=" ",
                        file=file,
                    )
                )
                file.unlink()

            except Exception as ex:
                LOG.error(f"Failed sending email report {file} with error: {ex}")

        if os.listdir(self.daily_report_path) == []:
            self.daily_report_path.rmdir()

        if os.listdir(self.daily_bundle_path) == []:
            self.daily_bundle_path.rmdir()

    def sync_files_sftp(self) -> None:
        self.check_username()
        transport = paramiko.Transport((self.config.fohm.host, self.config.fohm.port))
        ed_key = paramiko.Ed25519Key.from_private_key_file(self.config.fohm.key)
        transport.connect(username=self.config.fohm.username, pkey=ed_key)
        sftp = paramiko.SFTPClient.from_transport(transport)
        for file in self.daily_rawdata_path.iterdir():
            LOG.info(f"Sending {file} via SFTP, dry-run {self.dry_run}")
            if self._dry_run:
                continue

            try:
                sftp.put(file.as_posix(), f"/till-fohm/{file.name}")
                LOG.info(f"Finished sending {file}")
                file.unlink()
            except Exception as ex:
                LOG.error(f"Failed sending {file} with error: {ex}")

        sftp.close()
        transport.close()

        if os.listdir(self.daily_rawdata_path) == []:
            self.daily_rawdata_path.rmdir()

    def update_upload_started_at(self, case_id: str) -> None:
        """Update timestamp for cases which started being processed as batch."""
        if self._dry_run:
            return
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        case.analyses[0].upload_started_at = dt.datetime.now()
        self.status_db.session.commit()

    def update_uploaded_at(self, case_id: str) -> None:
        """Update timestamp for a case which uploaded successfully."""
        if self._dry_run:
            return
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        case.analyses[0].uploaded_at = dt.datetime.now()
        self.status_db.session.commit()

    def parse_and_write_complementary_report(self) -> list[FohmComplementaryReport]:
        """Parse and write a complementary report."""
        complementary_reports_raw: list[dict] = self.get_reports_contents(self.daily_reports_list)
        unique_complementary_reports_raw: list[dict] = remove_duplicate_dicts(
            complementary_reports_raw
        )
        sars_cov_complementary_reports: list[FohmComplementaryReport] = (
            self.validate_fohm_complementary_reports(unique_complementary_reports_raw)
        )
        sars_cov_complementary_reports: list[FohmComplementaryReport] = (
            self.get_sars_cov_complementary_reports(sars_cov_complementary_reports)
        )
        self.add_sample_internal_id_to_complementary_reports(sars_cov_complementary_reports)
        self.add_region_lab_to_reports(sars_cov_complementary_reports)
        self.create_complementary_report(sars_cov_complementary_reports)
        return sars_cov_complementary_reports

    def parse_and_write_pangolin_report(self) -> list[FohmPangolinReport]:
        """Create and write a Pangolin report."""
        pangolin_reports_raw: list[dict] = self.get_reports_contents(self.daily_pangolin_list)
        unique_pangolin_reports_raw: list[dict] = remove_duplicate_dicts(pangolin_reports_raw)
        sars_cov_pangolin_reports: list[FohmPangolinReport] = self.validate_fohm_pangolin_reports(
            unique_pangolin_reports_raw
        )
        sars_cov_pangolin_reports: list[FohmPangolinReport] = self.get_sars_cov_pangolin_reports(
            sars_cov_pangolin_reports
        )
        self.add_sample_internal_id_to_pangolin_reports(sars_cov_pangolin_reports)
        self.add_region_lab_to_reports(sars_cov_pangolin_reports)
        self.create_pangolin_report(sars_cov_pangolin_reports)
        return sars_cov_pangolin_reports

    def aggregate_delivery(self, case_ids: list[str]) -> None:
        """
        Aggregate and hardlink reports.
        args:
            case_ids: The internal ids for cases to aggregate.
        """
        self.set_cases_to_aggregate(case_ids)
        self.create_daily_delivery_folders()
        sars_cov_complementary_reports: list[FohmComplementaryReport] = (
            self.parse_and_write_complementary_report()
        )
        self.link_sample_raw_data_files(sars_cov_complementary_reports)
        sars_cov_pangolin_reports: list[FohmPangolinReport] = self.parse_and_write_pangolin_report()
        self.link_sample_raw_data_files(sars_cov_pangolin_reports)

    def create_and_write_complementary_report(self, cases: list[str]) -> None:
        """Create and write a complementary report."""
        self.set_cases_to_aggregate(cases)
        self.create_daily_delivery_folders()
        self.parse_and_write_complementary_report()



import logging
from datetime import datetime
from pathlib import Path
from typing import Any

from pydantic import ValidationInfo

from cg.constants import (
    BALSAMIC_ANALYSIS_TYPE,
    NA_FIELD,
    NO_FIELD,
    PRECISION,
    REPORT_SEX,
    YES_FIELD,
)
from cg.constants.constants import Workflow
from cg.constants.report import REPORT_QC_FLAG
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.constants.subject import Sex
from cg.models.delivery.delivery import DeliveryFile
from cg.models.orders.constants import OrderType

LOG = logging.getLogger(__name__)


def get_report_string(value: Any) -> str:
    """Return report adapted string."""
    return str(value) if value else NA_FIELD


def get_boolean_as_string(value: bool | None) -> str:
    """Return delivery report adapted string representation of a boolean."""
    if isinstance(value, bool):
        return YES_FIELD if value else NO_FIELD
    return NA_FIELD


def get_number_as_string(value: Any) -> str:
    """
    Return string representation of a number. If None is provided, then it returns N/A.

    Raises:
        ValueError: If the input value cannot be converted to a float.
    """
    try:
        result: str = str(round(float(value), PRECISION))
        return result
    except TypeError:
        return NA_FIELD
    except ValueError:
        LOG.error(f"Value {value} cannot be converted to float")
        raise


def get_float_as_percentage(value: float | None) -> str:
    """Return string percentage representation of a float value."""
    return get_number_as_string(value * 100) if value or isinstance(value, float) else NA_FIELD


def get_date_as_string(date: datetime | None) -> str:
    """Return the date string representation (year, month, day) of a datetime object."""
    return str(date.date()) if date else NA_FIELD


def get_list_as_string(value: list[str] | None) -> str:
    """Return list elements as comma separated individual string values."""
    return ", ".join(v for v in value) if value else NA_FIELD


def get_delivered_files_as_file_names(delivery_files: list[DeliveryFile] | None) -> list[str] | str:
    """Return a list of validated file names given a list of delivery files."""
    return [file.destination_path.name for file in delivery_files] if delivery_files else NA_FIELD


def get_path_as_string(file_path: str | None) -> str:
    """Return a report validated file name."""
    return Path(file_path).name if file_path and Path(file_path).is_file() else NA_FIELD


def get_sex_as_string(sex: Sex | None) -> str:
    """Return a report adapted sex."""
    return get_report_string(REPORT_SEX.get(sex))


def get_prep_category_as_string(prep_category: SeqLibraryPrepCategory | None) -> str:
    """Return a report validated prep category as string."""
    if prep_category == OrderType.RML:
        LOG.error("The delivery report generation does not support RML samples")
        raise ValueError
    return get_report_string(prep_category)


def get_analysis_type_as_string(analysis_type: str | None, info: ValidationInfo) -> str:
    """Return the analysis type as an accepted string value for the delivery report."""
    if analysis_type and Workflow.BALSAMIC in info.data.get("workflow"):
        analysis_type: str = BALSAMIC_ANALYSIS_TYPE.get(analysis_type)
    return get_report_string(analysis_type)


def get_initial_qc_as_string(qc_flag: bool | None) -> str:
    """Return a report adapted initial quality control flag."""
    return get_report_string(REPORT_QC_FLAG.get(qc_flag))



"""Test for the clean illumina runs cmd."""

import logging
from pathlib import Path
from unittest import mock

from click.testing import CliRunner

from cg.cli.clean import clean_illumina_runs
from cg.models.cg_config import CGConfig
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData


def test_clean_illumina_runs_cmd(
    cli_runner: CliRunner,
    clean_illumina_sequencing_runs_context: CGConfig,
    tmp_sequencing_run_to_clean: IlluminaRunDirectoryData,
    tmp_sequencing_run_to_clean_path: Path,
    tmp_sequencing_run_not_to_clean_path: Path,
    tmp_sample_sheet_clean_illumina_sequencing_run_path: Path,
    caplog,
):
    """Test the clean flow cells command."""
    # GIVEN a config with StatusDB and Housekeeper
    caplog.set_level(logging.DEBUG)
    # GIVEN one sequencing run that should be cleaned and one that should not
    assert tmp_sequencing_run_not_to_clean_path.exists()
    assert tmp_sequencing_run_to_clean_path.exists()
    assert tmp_sample_sheet_clean_illumina_sequencing_run_path.exists()
    assert clean_illumina_sequencing_runs_context.status_db.get_illumina_sequencing_run_by_device_internal_id(
        tmp_sequencing_run_to_clean.id
    )
    assert clean_illumina_sequencing_runs_context.housekeeper_api.get_latest_bundle_version(
        "HLYWYDSXX"
    )
    # WHEN running the clean flow cells cli command
    with mock.patch(
        "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.is_directory_older_than_21_days",
        return_value=True,
    ):
        result = cli_runner.invoke(clean_illumina_runs, obj=clean_illumina_sequencing_runs_context)

    # THEN assert it exits with success
    assert result.exit_code == 0

    # THEN the flow cell fulfilling all cleaning criteria is deleted
    assert not tmp_sequencing_run_to_clean_path.exists()
    assert "Successfully removed the directory and its contents" in caplog.text

    # THEN the flow cell not fulfilling all cleaning criteria is not deleted
    assert tmp_sequencing_run_not_to_clean_path.exists()
    assert "IlluminaCleanRunError" in caplog.text


def test_clean_illumina_runs_cmd_dry_run(
    cli_runner: CliRunner,
    clean_illumina_sequencing_runs_context: CGConfig,
    tmp_sequencing_run_to_clean_path: Path,
    tmp_sequencing_run_not_to_clean_path: Path,
    caplog,
):
    """Test the clean flow cells command using dry-run."""
    # GIVEN a config with StatusDB and Housekeeper
    caplog.set_level(logging.DEBUG)
    # GIVEN one flow cell that should be cleaned and one that should not
    assert tmp_sequencing_run_not_to_clean_path.exists()
    assert tmp_sequencing_run_to_clean_path.exists()
    # WHEN running the clean flow cells cli command
    with mock.patch(
        "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.is_directory_older_than_21_days",
        return_value=True,
    ):
        result = cli_runner.invoke(
            clean_illumina_runs,
            ["--dry-run"],
            obj=clean_illumina_sequencing_runs_context,
        )

    # THEN assert it exits with success
    assert result.exit_code == 0

    # THEN the directory to clean is not deleted
    assert tmp_sequencing_run_to_clean_path.exists()
    assert tmp_sequencing_run_not_to_clean_path.exists()
    assert "Would have removed" in caplog.text



"""Tests for the illumina flow cell filters in the store."""

from sqlalchemy.orm import Query

from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.data_transfer.models import IlluminaFlowCellDTO
from cg.store.filters.status_illumina_flow_cell_filters import (
    filter_illumina_flow_cell_by_internal_id,
)
from cg.store.models import IlluminaFlowCell
from cg.store.store import Store


def test_filter_illumina_flow_cell_by_internal_id(
    illumina_flow_cell_dto: IlluminaFlowCellDTO,
    novaseq_x_flow_cell_id: str,
    store_with_illumina_sequencing_data: Store,
    seven_canonical_flow_cells: list[IlluminaRunDirectoryData],
):
    # GIVEN a store with the canonical Illumina flow cells
    flow_cells: Query = store_with_illumina_sequencing_data._get_query(table=IlluminaFlowCell)
    number_of_flow_cells_in_store: int = len(seven_canonical_flow_cells)
    assert flow_cells.count() == number_of_flow_cells_in_store

    # WHEN filtering an Illumina flow cell by internal id
    filtered_flow_cells: Query = filter_illumina_flow_cell_by_internal_id(
        flow_cells=flow_cells, internal_id=novaseq_x_flow_cell_id
    )

    # THEN a query with the correct flow cell should be returned
    assert filtered_flow_cells.count() < number_of_flow_cells_in_store
    assert filtered_flow_cells.first().internal_id == novaseq_x_flow_cell_id



import logging
from functools import wraps
from pathlib import Path
from tempfile import TemporaryDirectory

from pydantic import ValidationError
from requests import ConnectionError, HTTPError
from requests.exceptions import MissingSchema

from cg.clients.freshdesk.exceptions import (
    FreshdeskAPIException,
    FreshdeskModelException,
)
from cg.constants.constants import FileFormat
from cg.io.controller import WriteFile

LOG = logging.getLogger(__name__)


def extract_error_detail(error):
    """Extract detailed error information from HTTPError."""
    if error.response is not None:
        try:
            return error.response.json()
        except ValueError:
            return error.response.text
    return None


def log_and_raise(exception_class, message, error):
    """Log the error and raise the appropriate exception."""
    LOG.error(message)
    raise exception_class(error) from error


def handle_client_errors(func):
    """Decorator to handle and log errors in Freshdesk client methods."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HTTPError as error:
            error_detail = extract_error_detail(error)
            log_and_raise(
                FreshdeskAPIException,
                f"Failed request to Freshdesk: {error} - Status code: "
                f"{error.response.status_code if error.response else 'N/A'}, Details: {error_detail}",
                error,
            )
        except (MissingSchema, ConnectionError) as error:
            log_and_raise(FreshdeskAPIException, f"Request to Freshdesk failed: {error}", error)
        except ValidationError as error:
            log_and_raise(
                FreshdeskModelException, f"Invalid response from Freshdesk: {error}", error
            )
        except ValueError as error:
            log_and_raise(FreshdeskAPIException, f"Operation failed: {error}", error)
        except Exception as error:
            log_and_raise(
                FreshdeskAPIException, f"Unexpected error in Freshdesk client: {error}", error
            )

    return wrapper


def prepare_attachments(attachments: list[Path]) -> list[tuple[str, tuple[str, bytes]]]:
    """Prepare the attachments for a request."""
    return [
        ("attachments[]", (attachment.name, open(attachment, "rb"))) for attachment in attachments
    ]


def create_temp_attachment_file(content: dict, file_name: Path) -> TemporaryDirectory:
    """Create a file-based attachment."""
    if content and file_name:
        directory = TemporaryDirectory()
        WriteFile.write_file_from_content(
            content=content,
            file_format=FileFormat.JSON,
            file_path=Path(directory.name, "order.json"),
        )
        return directory
    else:
        LOG.error("Content or file path is None. Cannot create file attachment.")
        raise ValueError("Both content and file path must be provided and cannot be None")



from pathlib import Path
from typing import Any

import pytest
from _pytest.logging import LogCaptureFixture
from pytest_mock import MockFixture

from cg.apps.demultiplex.sample_sheet.sample_models import IlluminaSampleIndexSetting
from cg.exc import HousekeeperFileMissingError
from cg.models.cg_config import CGConfig
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from tests.store_helpers import StoreHelpers

GET_FLOW_CELL_SAMPLES: str = "cg.apps.demultiplex.sample_sheet.api.get_flow_cell_samples"


def test_get_create_sample_sheet_hk_has_bcl2fastq_sample_sheet(
    sample_sheet_context_broken_flow_cells: CGConfig,
    tmp_flow_cell_with_bcl2fastq_sample_sheet: Path,
    sample_sheet_bcl2fastq_bundle_data: dict[str, Any],
    hiseq_x_single_index_bcl_convert_lims_samples: list[IlluminaSampleIndexSetting],
    helpers: StoreHelpers,
    mocker: MockFixture,
):
    """Test that a BCL2FASTQ sample sheet in Housekeeper is updated to BCLConvert."""
    # GIVEN a sample sheet context with a sample sheet API and a Housekeeper API
    sample_sheet_api = sample_sheet_context_broken_flow_cells.sample_sheet_api
    hk_api = sample_sheet_context_broken_flow_cells.housekeeper_api

    # GIVEN a a flow cell with a BCL2FASTQ sample sheet going to be updated to BCLConvert
    flow_cell = IlluminaRunDirectoryData(tmp_flow_cell_with_bcl2fastq_sample_sheet)

    # GIVEN that the sample sheet is in Housekeeper
    helpers.ensure_hk_bundle(store=hk_api, bundle_data=sample_sheet_bcl2fastq_bundle_data)
    sample_sheet_path_hk: Path = hk_api.get_sample_sheet_path(flow_cell.id)

    # GIVEN that the flow cell samples are in LIMS
    mocker.patch(
        GET_FLOW_CELL_SAMPLES,
        return_value=hiseq_x_single_index_bcl_convert_lims_samples,
    )

    # WHEN getting the sample sheet in BCLConvert format
    sample_sheet_api.get_or_create_sample_sheet(flow_cell.full_name)

    # THEN the sample sheet in Housekeeper is replaced with the BCLConvert sample sheet
    sample_sheet_api.validate_sample_sheet(sample_sheet_path_hk)

    # THEN the sample sheet in the flow cell is replaced with the BCLConvert sample sheet
    sample_sheet_api.validate_sample_sheet(flow_cell.sample_sheet_path)


def test_get_create_sample_sheet_flow_cell_has_bcl2fastq_sample_sheet(
    sample_sheet_context_broken_flow_cells: CGConfig,
    tmp_flow_cell_with_bcl2fastq_sample_sheet: Path,
    hiseq_x_single_index_bcl_convert_lims_samples: list[IlluminaSampleIndexSetting],
    mocker: MockFixture,
    caplog: LogCaptureFixture,
):
    """
    Test that a BCL2FASTQ sample sheet in the flow cell directory is updated to BCLConvert
    and added to Housekeeeper.
    """
    # GIVEN a sample sheet context with a sample sheet API and a Housekeeper API
    sample_sheet_api = sample_sheet_context_broken_flow_cells.sample_sheet_api
    hk_api = sample_sheet_context_broken_flow_cells.housekeeper_api

    # GIVEN a a flow cell with a BCL2FASTQ sample sheet going to be updated to BCLConvert
    flow_cell = IlluminaRunDirectoryData(tmp_flow_cell_with_bcl2fastq_sample_sheet)

    # GIVEN that the sample sheet is not in Housekeeper
    with pytest.raises(HousekeeperFileMissingError):
        hk_api.get_sample_sheet_path(flow_cell.id)
    assert f"Sample sheet file for flowcell {flow_cell.id} not found in Housekeeper!" in caplog.text

    # GIVEN that the flow cell samples are in LIMS
    mocker.patch(
        GET_FLOW_CELL_SAMPLES,
        return_value=hiseq_x_single_index_bcl_convert_lims_samples,
    )

    # WHEN getting the sample sheet in BCLConvert format
    sample_sheet_api.get_or_create_sample_sheet(flow_cell.full_name)

    # THEN the sample sheet has been added to Housekeeper and passes BCLConvert validation
    sample_sheet_path_hk: Path = hk_api.get_sample_sheet_path(flow_cell.id)
    sample_sheet_api.validate_sample_sheet(sample_sheet_path_hk)

    # THEN the sample sheet in the flow cell directory is replaced with the BCLConvert sample sheet
    sample_sheet_api.validate_sample_sheet(flow_cell.sample_sheet_path)



import logging
from datetime import datetime
from pathlib import Path
from typing import Iterator

from housekeeper.store.models import File, Version

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.constants import Workflow
from cg.constants.housekeeper_tags import WORKFLOW_PROTECTED_TAGS
from cg.store.models import Analysis
from cg.store.store import Store

LOG = logging.getLogger(__name__)


class CleanAPI:
    def __init__(self, status_db: Store, housekeeper_api: HousekeeperAPI):
        self.status_db = status_db
        self.housekeeper_api = housekeeper_api

    def get_bundle_files(self, before: datetime, workflow: Workflow) -> Iterator[list[File]]:
        """Get any bundle files for a specific version."""

        analysis: Analysis
        LOG.debug(
            f"number of {workflow} analyses before: {before} : {len(self.status_db.get_analyses_for_workflow_started_at_before(workflow=workflow, started_at_before=before))}"
        )

        for analysis in self.status_db.get_analyses_for_workflow_started_at_before(
            workflow=workflow, started_at_before=before
        ):
            bundle_name = analysis.case.internal_id

            hk_bundle_version: Version | None = self.housekeeper_api.version(
                bundle=bundle_name, date=analysis.started_at
            )
            if not hk_bundle_version:
                LOG.warning(
                    f"Version not found for "
                    f"bundle:{bundle_name}; "
                    f"workflow: {workflow}; "
                    f"date {analysis.started_at}"
                )
                continue

            LOG.info(
                f"Version found for "
                f"bundle:{bundle_name}; "
                f"workflow: {workflow}; "
                f"date {analysis.started_at}"
            )
            yield self.housekeeper_api.get_files(
                bundle=bundle_name, version=hk_bundle_version.id
            ).all()

    @staticmethod
    def has_protected_tags(file: File, protected_tags_lists: list[list[str]]) -> bool:
        """Check if a file has any protected tags"""

        LOG.info(f"File {file.full_path} has the tags {file.tags}")
        file_tags: list[str] = HousekeeperAPI.get_tag_names_from_file(file)

        _has_protected_tags: bool = False
        for protected_tags in protected_tags_lists:
            if set(protected_tags).issubset(set(file_tags)):
                LOG.debug(
                    f"File {file.full_path} has the protected tag(s) {protected_tags}, skipping."
                )
                _has_protected_tags = True
                break

        if not _has_protected_tags:
            LOG.info(f"File {file.full_path} has no protected tags.")
        return _has_protected_tags

    def get_unprotected_existing_bundle_files(self, before: datetime) -> Iterator[File]:
        """Returns all existing bundle files from analyses started before 'before' that have no protected tags"""

        workflow: Workflow
        for workflow in Workflow:
            protected_tags_lists = WORKFLOW_PROTECTED_TAGS.get(workflow)
            if not protected_tags_lists:
                LOG.debug(f"No protected tags defined for {workflow}, skipping")
                continue

            hk_files: list[File]
            for hk_files in self.get_bundle_files(before=before, workflow=workflow):
                hk_file: File
                for hk_file in hk_files:
                    if self.has_protected_tags(hk_file, protected_tags_lists=protected_tags_lists):
                        continue

                    file_path: Path = Path(hk_file.full_path)
                    if not file_path.exists():
                        LOG.info(f"File {file_path} not on disk.")
                        continue
                    LOG.info(f"File {file_path} found on disk.")
                    yield hk_file



from cg.models.orders.order import OrderIn
from cg.services.orders.submitters.order_submitter import ValidateOrderService
from cg.store.store import Store


class ValidateFastqOrderService(ValidateOrderService):

    def __init__(self, status_db: Store):
        self.status_db = status_db

    def validate_order(self, order: OrderIn) -> None:
        pass



"""Test CG CLI upload observations module."""

import logging

import pytest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.upload.observations import upload_observations_to_loqusdb
from cg.cli.upload.observations.utils import get_observations_api, get_observations_verified_case
from cg.constants import EXIT_SUCCESS
from cg.constants.constants import Workflow
from cg.constants.sequencing import SeqLibraryPrepCategory
from cg.constants.subject import PhenotypeStatus
from cg.exc import CaseNotFoundError
from cg.meta.observations.mip_dna_observations_api import MipDNAObservationsAPI
from cg.meta.observations.raredisease_observations_api import RarediseaseObservationsAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case, CaseSample, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_observations(
    cg_context: CGConfig, cli_runner: CliRunner, helpers: StoreHelpers, caplog: LogCaptureFixture
):
    """Test upload of observations."""
    caplog.set_level(logging.DEBUG)
    store: Store = cg_context.status_db

    # GIVEN a case ready to be uploaded to Loqusdb
    case: Case = helpers.add_case(store)
    case.customer.loqus_upload = True
    sample: Sample = helpers.add_sample(
        store, application_type=SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING
    )
    link: CaseSample = store.relate_sample(case=case, sample=sample, status=PhenotypeStatus.UNKNOWN)
    store.session.add(link)

    # WHEN trying to do a dry run upload to Loqusdb
    result = cli_runner.invoke(
        upload_observations_to_loqusdb, [case.internal_id, "--dry-run"], obj=cg_context
    )

    # THEN the execution should have been successful and stop at a dry run step
    assert result.exit_code == EXIT_SUCCESS
    assert "Dry run" in caplog.text


def test_get_observations_case(cg_context: CGConfig, helpers: StoreHelpers):
    """Test get observations supported case."""
    store: Store = cg_context.status_db

    # GIVEN an observations valid case
    case: Case = helpers.add_case(store)
    case.customer.loqus_upload = True

    # WHEN retrieving a case given a specific case ID
    extracted_case: Case = get_observations_verified_case(cg_context, case.internal_id, upload=True)

    # THEN the extracted case should match the stored one
    assert extracted_case == case


def test_get_observations_case_invalid_id(cg_context: CGConfig, caplog: LogCaptureFixture):
    """Test get observations case providing an incorrect ID."""
    caplog.set_level(logging.DEBUG)

    # WHEN retrieving a case given a specific case ID
    with pytest.raises(CaseNotFoundError):
        # THEN a CaseNotFoundError should be raised
        get_observations_verified_case(cg_context, "invalid_case_id", upload=True)

    assert "Invalid case ID. Retrieving available cases for Loqusdb actions." in caplog.text


def test_get_observations_case_to_upload(cg_context: CGConfig, helpers: StoreHelpers):
    """Test get case ready to be uploaded to Loqusdb."""
    store: Store = cg_context.status_db

    # GIVEN a case ready to be uploaded to Loqusdb
    case: Case = helpers.add_case(store)
    case.customer.loqus_upload = True

    # WHEN retrieving a case given a specific case ID
    extracted_case: Case = get_observations_verified_case(cg_context, case.internal_id, upload=True)

    # THEN the extracted case should match the stored one
    assert extracted_case == case


def test_get_observations_api(cg_context: CGConfig, helpers: StoreHelpers):
    """Test get observation API given a Loqusdb supported case."""
    store: Store = cg_context.status_db

    # GIVEN a Loqusdb supported case
    case: Case = helpers.add_case(store, data_analysis=Workflow.MIP_DNA)
    sample: Sample = helpers.add_sample(
        store, application_type=SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING
    )
    link: CaseSample = store.relate_sample(case=case, sample=sample, status=PhenotypeStatus.UNKNOWN)
    store.session.add(link)

    # WHEN retrieving the observation API
    observations_api: MipDNAObservationsAPI = get_observations_api(
        context=cg_context, case_id=case.internal_id, upload=True
    )

    # THEN a MIP-DNA API should be returned
    assert observations_api
    assert isinstance(observations_api, MipDNAObservationsAPI)


def test_get_observations_api_raredisease(cg_context: CGConfig, helpers: StoreHelpers):
    """Test get observation API given a Loqusdb supported case."""
    store: Store = cg_context.status_db

    # GIVEN a Loqusdb supported case
    case: Case = helpers.add_case(store, data_analysis=Workflow.RAREDISEASE)
    sample: Sample = helpers.add_sample(
        store, application_type=SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING
    )
    case_sample: CaseSample = store.relate_sample(
        case=case, sample=sample, status=PhenotypeStatus.UNKNOWN
    )

    store.session.add(case_sample)

    # WHEN retrieving the observation API
    observations_api: RarediseaseObservationsAPI = get_observations_api(
        context=cg_context, case_id=case.internal_id, upload=True
    )

    # THEN a RAREDISEASE ObservationAPI should be returned
    assert observations_api
    assert isinstance(observations_api, RarediseaseObservationsAPI)



"""Tests for the CleanFlowCellsAPI."""

from datetime import datetime
from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.meta.clean.clean_retrieved_spring_files import CleanRetrievedSpringFilesAPI


@pytest.fixture
def clean_retrieved_spring_files_api(
    real_housekeeper_api: HousekeeperAPI, tmp_path
) -> CleanRetrievedSpringFilesAPI:
    """Returns a CleanRetrievedSpringFilesAPI."""
    real_housekeeper_api.root_dir = tmp_path
    return CleanRetrievedSpringFilesAPI(housekeeper_api=real_housekeeper_api, dry_run=False)


@pytest.fixture
def path_to_old_retrieved_spring_file() -> str:
    return Path("path", "to", "old_retrieved_spring_file").as_posix()


@pytest.fixture
def path_to_newly_retrieved_spring_file() -> str:
    return Path("path", "to", "newly_retrieved_spring_file").as_posix()


@pytest.fixture
def path_to_archived_but_not_retrieved_spring_file() -> str:
    return Path("path", "to", "archived_spring_file").as_posix()


@pytest.fixture
def path_to_fastq_file() -> str:
    return Path("path", "to", "fastq_file").as_posix()


@pytest.fixture
def paths_for_populated_clean_retrieved_spring_files_api(
    path_to_old_retrieved_spring_file: str,
    path_to_newly_retrieved_spring_file: str,
    path_to_archived_but_not_retrieved_spring_file: str,
    path_to_fastq_file: str,
) -> list[str]:
    return [
        path_to_old_retrieved_spring_file,
        path_to_newly_retrieved_spring_file,
        path_to_archived_but_not_retrieved_spring_file,
        path_to_fastq_file,
    ]


@pytest.fixture
def retrieved_test_bundle_name() -> str:
    return "retrieved_test_bundle"


@pytest.fixture
def path_to_old_spring_file_in_housekeeper(
    retrieved_test_bundle_name: str, path_to_old_retrieved_spring_file
) -> str:
    return Path(
        retrieved_test_bundle_name,
        str(datetime.today().date()),
        Path(path_to_old_retrieved_spring_file).name,
    ).as_posix()


@pytest.fixture
def populated_clean_retrieved_spring_files_api(
    clean_retrieved_spring_files_api: CleanRetrievedSpringFilesAPI,
    paths_for_populated_clean_retrieved_spring_files_api: list[str],
    retrieved_test_bundle_name: str,
    archival_job_id_miria: int,
    retrieval_job_id_miria: int,
    timestamp: datetime,
    timestamp_yesterday: datetime,
    old_timestamp: datetime,
    tmp_path,
) -> CleanRetrievedSpringFilesAPI:
    """
    Returns a populated CleanRetrievedSpringFilesAPI, containing a bundle with one version and the following files:
        - an archived Spring file which has not been retrieved
        - an archived Spring file which was retrieved 1 day ago
        - an archived Spring file which was retrieved in the year 1900
        - a Fastq file
    """
    clean_retrieved_spring_files_api.housekeeper_api.add_bundle_and_version_if_non_existent(
        bundle_name=retrieved_test_bundle_name
    )
    clean_retrieved_spring_files_api.housekeeper_api.commit()
    for path in paths_for_populated_clean_retrieved_spring_files_api:
        tags: list[str] = (
            [SequencingFileTag.SPRING]
            if SequencingFileTag.SPRING in path
            else [SequencingFileTag.FASTQ]
        )
        Path(tmp_path, path).parent.mkdir(parents=True, exist_ok=True)
        file_to_add = Path(tmp_path, path)
        file_to_add.touch()
        clean_retrieved_spring_files_api.housekeeper_api.add_and_include_file_to_latest_version(
            file=file_to_add, bundle_name=retrieved_test_bundle_name, tags=tags
        )
    for file in clean_retrieved_spring_files_api.housekeeper_api.get_files(
        bundle=retrieved_test_bundle_name
    ):
        Path(file.full_path).parent.mkdir(parents=True, exist_ok=True)
        Path(file.full_path).touch()
        if "spring" in file.path:
            clean_retrieved_spring_files_api.housekeeper_api.add_archives(
                files=[file], archive_task_id=archival_job_id_miria
            )
            file.archive.archived_at = timestamp
            if "retrieved" in file.path:
                file.archive.retrieval_task_id = retrieval_job_id_miria
                file.archive.retrieved_at = (
                    old_timestamp if "old" in file.path else timestamp_yesterday
                )
            clean_retrieved_spring_files_api.housekeeper_api.add_commit(file.archive)

    clean_retrieved_spring_files_api.housekeeper_api.commit()
    return clean_retrieved_spring_files_api



import logging
from datetime import datetime
from pathlib import Path

from cg.apps.housekeeper import models as hk_models
from cg.io.json import read_json_stream, write_json_stream
from cg.utils.commands import Process

from .models import CGDeliverables

LOG = logging.getLogger(__name__)


class HermesApi:
    """Class to communicate with hermes"""

    def __init__(self, config: dict):
        self.process = Process(
            binary=config["hermes"]["binary_path"],
        )
        self.container_mount_volume = config["hermes"]["container_mount_volume"]

    def convert_deliverables(
        self,
        deliverables_file: Path,
        workflow: str,
        analysis_type: str | None = None,
        force: bool = False,
    ) -> CGDeliverables:
        """Convert deliverables file in raw workflow format to CG format with Hermes."""
        LOG.info("Converting workflow deliverables to CG deliverables")
        convert_command = [
            "run",
            "--bind",
            self.container_mount_volume,
            "/home/proj/stage/singularity_containers/hermes_latest.sif",
            "convert",
            "deliverables",
            "--workflow",
            workflow,
            str(deliverables_file),
        ]
        if analysis_type:
            convert_command.extend(["--analysis-type", analysis_type])
        if force:
            convert_command.append("--force")
        self.process.run_command(convert_command)
        json_stream: str = self.process.stdout
        if force:
            data: dict[str, any] = read_json_stream(json_stream)
            for file_tag in data["files"]:
                file_tag["mandatory"] = False
            json_stream: str = write_json_stream(data)
        return CGDeliverables.model_validate_json(json_stream)

    def create_housekeeper_bundle(
        self,
        bundle_name: str,
        deliverables: Path,
        workflow: str,
        analysis_type: str | None,
        created: datetime | None,
        force: bool = False,
    ) -> hk_models.InputBundle:
        """Convert workflow deliverables to a Housekeeper bundle ready to be inserted into Housekeeper."""
        cg_deliverables: CGDeliverables = self.convert_deliverables(
            deliverables_file=deliverables,
            workflow=workflow,
            analysis_type=analysis_type,
            force=force,
        )
        return self.get_housekeeper_bundle(
            deliverables=cg_deliverables, created=created, bundle_name=bundle_name
        )

    @staticmethod
    def get_housekeeper_bundle(
        deliverables: CGDeliverables, bundle_name: str, created: datetime | None = None
    ) -> hk_models.InputBundle:
        """Convert a deliverables object to a housekeeper object"""
        bundle_info = {
            "name": bundle_name,
            "files": [file_info.model_dump() for file_info in deliverables.files],
        }
        if created:
            bundle_info["created"] = created

        return hk_models.InputBundle(**bundle_info)



from cg.services.delivery_message.messages.analysis_scout_message import (
    AnalysisScoutMessage,
)
from cg.services.delivery_message.messages.covid_message import CovidMessage
from cg.services.delivery_message.messages.fastq_analysis_scout_message import (
    FastqAnalysisScoutMessage,
)
from cg.services.delivery_message.messages.fastq_message import FastqMessage
from cg.services.delivery_message.messages.fastq_scout_message import FastqScoutMessage
from cg.services.delivery_message.messages.microsalt_mwr_message import (
    MicrosaltMwrMessage,
)
from cg.services.delivery_message.messages.microsalt_mwx_message import (
    MicrosaltMwxMessage,
)
from cg.services.delivery_message.messages.scout_message import ScoutMessage
from cg.services.delivery_message.messages.statina_message import StatinaMessage
from cg.services.delivery_message.messages.rna_delivery_message import (
    RNAScoutStrategy,
    RNAFastqStrategy,
    RNAAnalysisStrategy,
    RNAFastqAnalysisStrategy,
    RNAUploadMessageStrategy,
    RNADeliveryMessage,
)



import os
from unittest.mock import Mock
import pytest
from pathlib import Path

from cg.services.deliver_files.file_formatter.files.abstract import FileFormatter
from cg.services.deliver_files.file_formatter.files.mutant_service import (
    MutantFileFormatter,
)
from cg.services.fastq_concatenation_service.fastq_concatenation_service import (
    FastqConcatenationService,
)
from cg.services.deliver_files.file_fetcher.models import (
    CaseFile,
    SampleFile,
)
from cg.services.deliver_files.file_formatter.destination.models import FormattedFile
from cg.services.deliver_files.file_formatter.files.case_service import (
    CaseFileFormatter,
)
from cg.services.deliver_files.file_formatter.files.concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.files.sample_service import (
    SampleFileFormatter,
    FileManager,
)
from cg.services.deliver_files.file_formatter.path_name.flat_structure import (
    FlatStructurePathFormatter,
)
from cg.services.deliver_files.file_formatter.path_name.nested_structure import (
    NestedStructurePathFormatter,
)


@pytest.mark.parametrize(
    "moved_files,expected_formatted_files,file_formatter",
    [
        (
            "expected_moved_analysis_case_delivery_files",
            "expected_formatted_analysis_case_files",
            CaseFileFormatter(
                file_manager=FileManager(), path_name_formatter=NestedStructurePathFormatter()
            ),
        ),
        (
            "expected_moved_analysis_sample_delivery_files",
            "expected_formatted_analysis_sample_files",
            SampleFileFormatter(
                file_manager=FileManager(), path_name_formatter=NestedStructurePathFormatter()
            ),
        ),
        (
            "fastq_concatenation_sample_files",
            "expected_concatenated_fastq_formatted_files",
            SampleFileConcatenationFormatter(
                file_manager=FileManager(),
                path_name_formatter=NestedStructurePathFormatter(),
                concatenation_service=FastqConcatenationService(),
            ),
        ),
        (
            "fastq_concatenation_sample_files_flat",
            "expected_concatenated_fastq_flat_formatted_files",
            SampleFileConcatenationFormatter(
                file_manager=FileManager(),
                path_name_formatter=FlatStructurePathFormatter(),
                concatenation_service=FastqConcatenationService(),
            ),
        ),
    ],
)
def test_file_formatters(
    moved_files: list[CaseFile | SampleFile],
    expected_formatted_files: list[FormattedFile],
    file_formatter: FileFormatter,
    request,
):
    # GIVEN existing case files, a case file formatter and a ticket directory path and a customer inbox
    moved_files: list[CaseFile | SampleFile] = request.getfixturevalue(moved_files)
    expected_formatted_files: list[FormattedFile] = request.getfixturevalue(
        expected_formatted_files
    )
    delivery_path: Path = moved_files[0].file_path.parent

    os.makedirs(delivery_path, exist_ok=True)

    for moved_file in moved_files:
        moved_file.file_path.touch()

    # WHEN formatting the case files
    formatted_files: list[FormattedFile] = file_formatter.format_files(
        moved_files=moved_files,
        delivery_path=delivery_path,
    )

    # THEN the case files should be formatted
    assert formatted_files == expected_formatted_files
    for file in formatted_files:
        assert file.formatted_path.exists()
        assert not file.original_path.exists()


def test_mutant_file_formatter(
    mutant_moved_files: list[SampleFile],
    expected_mutant_formatted_files: list[FormattedFile],
    lims_naming_metadata: str,
):
    # GIVEN existing ticket directory path and a customer inbox
    ticket_dir_path: Path = mutant_moved_files[0].file_path.parent

    os.makedirs(ticket_dir_path, exist_ok=True)

    for moved_file in mutant_moved_files:
        moved_file.file_path.touch()

    lims_mock = Mock()
    lims_mock.get_sample_region_and_lab_code.return_value = lims_naming_metadata
    file_formatter = MutantFileFormatter(
        file_manager=FileManager(),
        file_formatter=SampleFileConcatenationFormatter(
            file_manager=FileManager(),
            path_name_formatter=NestedStructurePathFormatter(),
            concatenation_service=FastqConcatenationService(),
        ),
        lims_api=lims_mock,
    )

    # WHEN formatting the files
    formatted_files: list[FormattedFile] = file_formatter.format_files(
        moved_files=mutant_moved_files,
        delivery_path=ticket_dir_path,
    )

    # THEN the files should be formatted
    assert formatted_files == expected_mutant_formatted_files
    for file in formatted_files:
        assert file.formatted_path.exists()
        assert not file.original_path.exists()


def test_concatenation_sample_name_match():
    # GIVEN a concatenation service and a list of file paths and a sample name that is a number
    sample_name = "12"
    concatentation_formatter = SampleFileConcatenationFormatter(
        file_manager=Mock(),
        path_name_formatter=Mock(),
        concatenation_service=Mock(),
    )
    # GIVEN two sets of file paths that should match and not match the sample name
    should_match_file_paths = [
        Path("path/to/FC_12_L001_R1_001.fastq.gz"),
        Path("path/to/FC_12_L002_R1_001.fastq.gz"),
        Path("path/to/FC_12_L001_R2_001.fastq.gz"),
        Path("path/to/FC_12_L002_R2_001.fastq.gz"),
    ]
    should_not_match_file_paths = [
        Path("path/to/FC_123_L001_R1_001.fastq.gz"),
        Path("path/to/FC_123_L002_R1_001.fastq.gz"),
        Path("path/to/FC_123_L001_R2_001.fastq.gz"),
        Path("path/to/FC_123_L002_R2_001.fastq.gz"),
    ]

    # WHEN checking if the file paths match the sample name

    # THEN the file paths that should match should return True and the file paths that should not match should return False
    for file_path in should_match_file_paths:
        assert (
            concatentation_formatter._has_expected_sample_name_format_match(
                file_path=file_path, sample_name=sample_name
            )
            is True
        )

    for file_path in should_not_match_file_paths:
        assert (
            concatentation_formatter._has_expected_sample_name_format_match(
                file_path=file_path, sample_name=sample_name
            )
            is False
        )



"""Tests for the meta BackupAPI."""

import fnmatch
import logging
from typing import Callable

import mock
import pytest

from cg.constants import EXIT_FAIL, SequencingRunDataAvailability
from cg.exc import (
    DsmcAlreadyRunningError,
    IlluminaRunAlreadyBackedUpError,
    IlluminaRunEncryptionError,
    PdcError,
)
from cg.models.cg_config import CGConfig, PDCArchivingDirectory
from cg.services.illumina.backup.backup_service import IlluminaBackupService
from cg.services.illumina.backup.encrypt_service import IlluminaRunEncryptionService
from cg.services.illumina.backup.utils import (
    get_latest_archived_encryption_key_path,
    get_latest_archived_sequencing_run_path,
)
from cg.services.pdc_service.pdc_service import PdcService
from cg.store.models import IlluminaSequencingRun
from cg.store.store import Store
from tests.conftest import create_process_response


@pytest.mark.parametrize(
    "flow_cell_name",
    ["new_flow_cell", "old_flow_cell", "ancient_flow_cell"],
)
def test_query_pdc_for_run(
    caplog,
    flow_cell_name: str,
    mock_pdc_query_method: Callable,
    pdc_archiving_directory: PDCArchivingDirectory,
):
    """Tests query PDC for an Illumina run with a mock PDC query."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a Backup API
    backup_api = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=pdc_archiving_directory,
        status_db=mock.Mock(),
        tar_api=mock.Mock(),
        pdc_service=mock.Mock(),
        sequencing_runs_dir=mock.Mock(),
    )
    # GIVEN a mock pdc query method
    backup_api.pdc.query_pdc = mock_pdc_query_method

    # WHEN querying pdc for a run by the flow cell
    backup_api.query_pdc_for_sequencing_run(flow_cell_id=flow_cell_name)

    # THEN the flow cell is logged as found for one of the search patterns
    assert fnmatch.filter(
        names=caplog.messages, pat=f"Found archived files for PDC query:*{flow_cell_name}*.gpg"
    )


def test_maximum_processing_queue_full(store_with_illumina_sequencing_data: Store):
    # GIVEN that a sequencing run needs to be retrieved from PDC
    sequencing_runs: list[IlluminaSequencingRun] = store_with_illumina_sequencing_data._get_query(
        IlluminaSequencingRun
    ).all()
    for sequencing_run in sequencing_runs:
        sequencing_run.data_availability = SequencingRunDataAvailability.PROCESSING

    backup_api = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=mock.Mock(),
        status_db=store_with_illumina_sequencing_data,
        tar_api=mock.Mock(),
        pdc_service=mock.Mock(),
        sequencing_runs_dir=mock.Mock(),
    )

    # WHEN there's already a run being retrieved from PDC

    # THEN this method should return False
    assert backup_api.has_processing_queue_capacity() is False


def test_maximum_processing_queue_not_full(store_with_illumina_sequencing_data: Store):
    # GIVEN a store with a requested sequencing run
    sequencing_runs: list[IlluminaSequencingRun] = store_with_illumina_sequencing_data._get_query(
        IlluminaSequencingRun
    ).all()
    for sequencing_run in sequencing_runs:
        sequencing_run.data_availability = SequencingRunDataAvailability.REQUESTED

    backup_api = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=mock.Mock(),
        status_db=store_with_illumina_sequencing_data,
        tar_api=mock.Mock(),
        pdc_service=mock.Mock(),
        sequencing_runs_dir=mock.Mock(),
    )
    # WHEN there are no runs being retrieved from PDC

    # THEN this method should return True
    assert backup_api.has_processing_queue_capacity()


def test_get_first_run_next_requested(
    store_with_illumina_sequencing_data: Store, novaseq_x_flow_cell_id
):
    # GIVEN a store with a requested sequencing run
    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    sequencing_run.data_availability = SequencingRunDataAvailability.REQUESTED

    backup_api = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=mock.Mock(),
        status_db=store_with_illumina_sequencing_data,
        tar_api=mock.Mock(),
        pdc_service=mock.Mock(),
        sequencing_runs_dir=mock.Mock(),
    )

    # WHEN a sequencing is requested to be retrieved from PDC
    requested_run = backup_api.get_first_run()

    # THEN a sequencing run is returned
    assert requested_run == sequencing_run


def test_get_first_run_no_run_requested(store_with_illumina_sequencing_data: Store):
    # GIVEN a store with sequencing runs that are not requested
    sequencing_runs: list[IlluminaSequencingRun] = store_with_illumina_sequencing_data._get_query(
        IlluminaSequencingRun
    ).all()
    for sequencing_run in sequencing_runs:
        sequencing_run.data_availability = SequencingRunDataAvailability.ON_DISK

    backup_api = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=mock.Mock(),
        status_db=store_with_illumina_sequencing_data,
        tar_api=mock.Mock(),
        pdc_service=mock.Mock(),
        sequencing_runs_dir=mock.Mock(),
    )

    # WHEN getting the first requested run
    requested_run = backup_api.get_first_run()

    # THEN no sequencing run is returned
    assert requested_run is None


@mock.patch(
    "cg.services.illumina.backup.backup_service.IlluminaBackupService.has_processing_queue_capacity"
)
@mock.patch("cg.store.models.IlluminaSequencingRun")
def test_fetch_sequencing_run_processing_queue_full(
    mock_sequencing_run, mock_check_processing, caplog
):
    caplog.set_level(logging.INFO)

    # GIVEN we check if an Illumina run needs to be retrieved from PDC
    backup_api = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=mock.Mock(),
        status_db=mock.Mock(),
        tar_api=mock.Mock(),
        pdc_service=mock.Mock(),
        sequencing_runs_dir=mock.Mock(),
    )

    # WHEN the processing queue is full
    backup_api.has_processing_queue_capacity.return_value = False
    result = backup_api.fetch_sequencing_run(mock_sequencing_run)

    # THEN no sequencing run will be fetched and a log message indicates that the processing queue is
    # full
    assert result is None
    assert "Processing queue is full" in caplog.text


@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.get_first_run")
@mock.patch(
    "cg.services.illumina.backup.backup_service.IlluminaBackupService.has_processing_queue_capacity"
)
@mock.patch("cg.store")
def test_fetch_sequencing_run_no_runs_requested(
    mock_store,
    has_processing_queue_capacity,
    mock_get_first_run,
    caplog,
):
    """Tests the fetch sequencing run method of the backup API when no flow cell requested"""

    caplog.set_level(logging.INFO)

    # GIVEN we check if a flow cell needs to be retrieved from PDC
    backup_api = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=mock.Mock(),
        status_db=mock_store,
        tar_api=mock.Mock(),
        pdc_service=mock.Mock(),
        sequencing_runs_dir=mock.Mock(),
    )

    # WHEN no runs are requested
    backup_api.get_first_run.return_value = None
    backup_api.has_processing_queue_capacity.return_value = True

    # AND no flow cell has been specified
    result = backup_api.fetch_sequencing_run(None)

    # THEN no sequencing run will be fetched and a log message indicates that no runs have been requested
    assert result is None
    assert "No sequencing run requested" in caplog.text


@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.unlink_files")
@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.create_rta_complete")
@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.create_copy_complete")
@mock.patch("cg.services.illumina.backup.backup_service.get_latest_archived_sequencing_run_path")
@mock.patch("cg.services.illumina.backup.backup_service.get_latest_archived_encryption_key_path")
@mock.patch(
    "cg.services.illumina.backup.backup_service.IlluminaBackupService.has_processing_queue_capacity"
)
def test_fetch_sequencing_run_retrieve_next_run(
    mock_get_latest_archived_encryption_key_path,
    mock_get_latest_archived_sequencing_run_path,
    mock_create_copy_complete,
    mock_create_rta_complete,
    archived_key,
    archived_illumina_run,
    cg_context,
    store_with_illumina_sequencing_data: Store,
    novaseq_x_flow_cell_id: str,
    caplog,
):
    """Tests the fetch sequencing run method of the backup API when retrieving next sequencing run"""

    caplog.set_level(logging.INFO)

    # GIVEN we check if a sequencing run needs to be retrieved from PDC
    backup_api = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=cg_context.illumina_backup_service.pdc_archiving_directory,
        status_db=store_with_illumina_sequencing_data,
        tar_api=mock.Mock(),
        pdc_service=mock.Mock(),
        sequencing_runs_dir="cg_context.flow_cells_dir",
    )

    # WHEN no sequencing run is specified, but a sequencing run in status-db has the status "requested"
    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    sequencing_run.data_availability = SequencingRunDataAvailability.REQUESTED
    backup_api.has_processing_queue_capacity.return_value = True
    get_latest_archived_encryption_key_path.return_value = archived_key
    get_latest_archived_sequencing_run_path.return_value = archived_illumina_run
    backup_api.tar_api.run_tar_command.return_value = None
    result = backup_api.fetch_sequencing_run(sequencing_run=None)

    # THEN the process to retrieve the sequencing run from PDC is started
    assert "retrieving from PDC" in caplog.text

    # AND when done the status of that sequencing run is set to "retrieved"
    assert sequencing_run.data_availability == SequencingRunDataAvailability.RETRIEVED

    # AND the elapsed time of the retrieval process is returned
    assert result > 0


@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.unlink_files")
@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.create_rta_complete")
@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.create_copy_complete")
@mock.patch("cg.services.illumina.backup.backup_service.get_latest_archived_sequencing_run_path")
@mock.patch("cg.services.illumina.backup.backup_service.get_latest_archived_encryption_key_path")
@mock.patch(
    "cg.services.illumina.backup.backup_service.IlluminaBackupService.has_processing_queue_capacity"
)
@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.get_first_run")
def test_fetch_sequencing_run_retrieve_specified_run(
    mock_get_first_run,
    mock_has_processing_queue_capacity,
    mock_get_latest_archived_encryption_key_path,
    mock_get_latest_archived_sequencing_run_path,
    mock_create_copy_complete,
    mock_create_rta_complete,
    mock_unlink_files,
    archived_key,
    archived_illumina_run,
    cg_context,
    dsmc_q_archive_output,
    store_with_illumina_sequencing_data: Store,
    novaseq_x_flow_cell_id: str,
    caplog,
):
    """Tests the fetch sequencing run method of the backup API when given a sequencing run."""

    caplog.set_level(logging.INFO)

    # GIVEN we want to retrieve a specific sequencing run from PDC
    backup_api = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=cg_context.illumina_backup_service.pdc_archiving_directory,
        status_db=store_with_illumina_sequencing_data,
        tar_api=mock.Mock(),
        pdc_service=mock.Mock(),
        sequencing_runs_dir="cg_context.flow_cells_dir",
    )
    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    sequencing_run.data_availability = SequencingRunDataAvailability.REQUESTED
    backup_api.has_processing_queue_capacity.return_value = True
    get_latest_archived_encryption_key_path.return_value = archived_key
    get_latest_archived_sequencing_run_path.return_value = archived_illumina_run
    backup_api.tar_api.run_tar_command.return_value = None
    result = backup_api.fetch_sequencing_run(sequencing_run=sequencing_run)

    # THEN no sequencing run is taken form statusdb
    mock_get_first_run.assert_not_called()

    # THEN the process to retrieve the sequencing run from PDC is started
    assert "retrieving from PDC" in caplog.text

    # AND when done the data availability of sequencing run is set to "retrieved"
    assert sequencing_run.data_availability == SequencingRunDataAvailability.RETRIEVED

    # AND the elapsed time of the retrieval process is returned
    assert result > 0


@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.unlink_files")
@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.create_rta_complete")
@mock.patch("cg.services.illumina.backup.backup_service.IlluminaBackupService.create_copy_complete")
@mock.patch(
    "cg.services.illumina.backup.backup_service.IlluminaBackupService.query_pdc_for_sequencing_run"
)
@mock.patch("cg.services.illumina.backup.backup_service.get_latest_archived_sequencing_run_path")
@mock.patch("cg.services.illumina.backup.backup_service.get_latest_archived_encryption_key_path")
def test_fetch_sequencing_run_integration(
    mock_sequencing_run_path,
    mock_key_path,
    mock_query,
    mock_create_copy_complete,
    mock_create_rta_complete,
    mock_unlink_files,
    cg_context,
    store_with_illumina_sequencing_data: Store,
    novaseq_x_flow_cell_id: str,
    dsmc_q_archive_output,
    caplog,
):
    """Component integration test for the BackupAPI, fetching a specified sequencing run."""

    caplog.set_level(logging.INFO)

    # GIVEN we want to retrieve a specific sequencing run from PDC
    backup_api = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=cg_context.illumina_backup_service.pdc_archiving_directory,
        status_db=store_with_illumina_sequencing_data,
        tar_api=mock.Mock(),
        pdc_service=mock.Mock(),
        sequencing_runs_dir=cg_context.run_instruments.illumina.sequencing_runs_dir,
    )
    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    sequencing_run.data_availability = SequencingRunDataAvailability.REQUESTED
    mock_query.return_value = dsmc_q_archive_output

    backup_api.tar_api.run_tar_command.return_value = None
    result = backup_api.fetch_sequencing_run(sequencing_run=sequencing_run)

    # THEN the process to retrieve the sequencing run from PDC is started
    assert "retrieving from PDC" in caplog.text

    # AND when done the status of that sequencing run is set to "retrieved"
    assert sequencing_run.data_availability == SequencingRunDataAvailability.RETRIEVED

    # AND the elapsed time of the retrieval process is returned
    assert result > 0


def test_validate_is_sequencing_run_backup_possible(
    base_store: Store,
    caplog,
    cg_context: CGConfig,
    store_with_illumina_sequencing_data: Store,
    novaseq_x_flow_cell_id: str,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
):
    """Tests checking if a back-up of an Illumina run is possible."""
    caplog.set_level(logging.DEBUG)
    # GIVEN an instance of the PDC API
    pdc_service = cg_context.pdc_service

    # GIVEN a backup service
    backup_service = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=cg_context.illumina_backup_service.pdc_archiving_directory,
        status_db=base_store,
        tar_api=mock.Mock(),
        pdc_service=pdc_service,
        sequencing_runs_dir=cg_context.run_instruments.illumina.sequencing_runs_dir,
    )

    # GIVEN no Dsmc process is running

    # GIVEN a sequencing run which is not backed up
    cg_context.status_db_ = store_with_illumina_sequencing_data
    sequencing_run: IlluminaSequencingRun = (
        cg_context.status_db_.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    assert not sequencing_run.has_backup

    # GIVEN that encryption is completed
    illumina_run_encryption_service.run_encryption_dir.mkdir(parents=True)
    illumina_run_encryption_service.complete_file_path.touch()

    # WHEN checking if back-up is possible
    backup_service.validate_is_run_backup_possible(
        sequencing_run=sequencing_run,
        illumina_run_encryption_service=illumina_run_encryption_service,
    )

    # THEN communicate that it passed
    assert "Sequencing run can be backed up" in caplog.text


def test_validate_is_flow_cell_backup_when_dsmc_is_already_running(
    base_store: Store,
    cg_context: CGConfig,
    novaseq_x_flow_cell_id: str,
    store_with_illumina_sequencing_data: Store,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
    mocker,
):
    """Tests checking if a back-up of a sequencing run is possible when Dsmc is already running."""
    # GIVEN an instance of the PDC API
    pdc_service = cg_context.pdc_service

    # GIVEN a backup service
    backup_service = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=cg_context.illumina_backup_service.pdc_archiving_directory,
        status_db=base_store,
        tar_api=mock.Mock(),
        pdc_service=pdc_service,
        sequencing_runs_dir=cg_context.run_instruments.illumina.sequencing_runs_dir,
    )

    # GIVEN a Dsmc process is already running
    mocker.patch.object(PdcService, "validate_is_dsmc_running", return_value=True)

    # GIVEN a sequencing run which is not backed up
    cg_context.status_db_ = store_with_illumina_sequencing_data
    sequencing_run: IlluminaSequencingRun = (
        cg_context.status_db_.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    assert not sequencing_run.has_backup

    # WHEN checking if back-up is possible
    with pytest.raises(DsmcAlreadyRunningError):
        backup_service.validate_is_run_backup_possible(
            sequencing_run=sequencing_run,
            illumina_run_encryption_service=illumina_run_encryption_service,
        )

        # THEN error should be raised


def test_validate_is_illumina_run_backup_when_already_backed_up(
    base_store: Store,
    cg_context: CGConfig,
    store_with_illumina_sequencing_data: Store,
    novaseq_x_flow_cell_id: str,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
):
    """Tests checking if a back-up of ab Illumina run is possible when the run is already backed up."""
    # GIVEN an instance of the PDC API
    pdc_service = cg_context.pdc_service

    # GIVEN a backup service
    backup_service = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=cg_context.illumina_backup_service.pdc_archiving_directory,
        status_db=base_store,
        tar_api=mock.Mock(),
        pdc_service=pdc_service,
        sequencing_runs_dir=cg_context.run_instruments.illumina.sequencing_runs_dir,
    )

    # GIVEN a sequencing run which backed up
    cg_context.status_db_ = store_with_illumina_sequencing_data
    sequencing_run: IlluminaSequencingRun = (
        cg_context.status_db_.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    sequencing_run.has_backup = True

    # WHEN checking if back-up is possible
    with pytest.raises(IlluminaRunAlreadyBackedUpError):
        backup_service.validate_is_run_backup_possible(
            sequencing_run=sequencing_run,
            illumina_run_encryption_service=illumina_run_encryption_service,
        )

        # THEN error should be raised


def test_validate_is_run_backup_when_encryption_is_not_complete(
    base_store: Store,
    cg_context: CGConfig,
    store_with_illumina_sequencing_data: Store,
    novaseq_x_flow_cell_id: str,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
):
    """Tests checking if a back-up of an Illumina run is possible when encryption is not complete."""
    # GIVEN an instance of the PDC API
    pdc_service = cg_context.pdc_service

    # GIVEN a backup service
    backup_service = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=cg_context.illumina_backup_service.pdc_archiving_directory,
        status_db=base_store,
        tar_api=mock.Mock(),
        pdc_service=pdc_service,
        sequencing_runs_dir=cg_context.run_instruments.illumina.sequencing_runs_dir,
    )

    # GIVEN a sequencing run which is not backed up
    cg_context.status_db_ = store_with_illumina_sequencing_data
    sequencing_run: IlluminaSequencingRun = (
        cg_context.status_db_.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    assert not sequencing_run.has_backup

    # WHEN checking if back-up is possible
    with pytest.raises(IlluminaRunEncryptionError):
        backup_service.validate_is_run_backup_possible(
            sequencing_run=sequencing_run,
            illumina_run_encryption_service=illumina_run_encryption_service,
        )

        # THEN error should be raised


def test_backup_illumina_runs(
    base_store: Store,
    cg_context: CGConfig,
    store_with_illumina_sequencing_data: Store,
    novaseq_x_flow_cell_id: str,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
    mocker,
):
    """Tests back-up Illumina runs."""
    # GIVEN an instance of the PDC API
    pdc_service = cg_context.pdc_service

    # GIVEN a backup service
    backup_service = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=cg_context.illumina_backup_service.pdc_archiving_directory,
        status_db=base_store,
        tar_api=mock.Mock(),
        pdc_service=pdc_service,
        sequencing_runs_dir=cg_context.run_instruments.illumina.sequencing_runs_dir,
    )
    # GIVEN a mocked archiving call
    mocker.patch.object(PdcService, "archive_file_to_pdc", return_value=None)

    # GIVEN a sequencing run which is not backed up
    cg_context.status_db_ = store_with_illumina_sequencing_data
    sequencing_run: IlluminaSequencingRun = (
        store_with_illumina_sequencing_data.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    assert not sequencing_run.has_backup

    # WHEN backing up the sequencing run
    backup_service.backup_run(
        files_to_archive=[
            illumina_run_encryption_service.final_passphrase_file_path,
            illumina_run_encryption_service.encrypted_gpg_file_path,
        ],
        store=base_store,
        sequencing_run=sequencing_run,
    )

    # THEN sequencing run should hava a back-up
    assert sequencing_run.has_backup


def test_backup_illumina_run_when_unable_to_archive(
    base_store: Store,
    cg_context: CGConfig,
    novaseq_x_flow_cell_id: str,
    illumina_run_encryption_service: IlluminaRunEncryptionService,
    store_with_illumina_sequencing_data: Store,
    caplog,
):
    """Tests back-up Illumina run when unable to archive."""
    caplog.set_level(logging.DEBUG)

    # GIVEN an instance of the PDC API
    pdc_service = cg_context.pdc_service

    # GIVEN a backup service
    backup_service = IlluminaBackupService(
        encryption_api=mock.Mock(),
        pdc_archiving_directory=cg_context.illumina_backup_service.pdc_archiving_directory,
        status_db=base_store,
        tar_api=mock.Mock(),
        pdc_service=pdc_service,
        sequencing_runs_dir=cg_context.run_instruments.illumina.sequencing_runs_dir,
    )
    # GIVEN a sequencing run which is not backed up
    cg_context.status_db_ = store_with_illumina_sequencing_data
    sequencing_run: IlluminaSequencingRun = (
        cg_context.status_db.get_illumina_sequencing_run_by_device_internal_id(
            novaseq_x_flow_cell_id
        )
    )
    assert not sequencing_run.has_backup

    # GIVEN that archiving fails
    with mock.patch(
        "cg.utils.commands.subprocess.run",
        return_value=create_process_response(return_code=EXIT_FAIL),
    ):
        # WHEN backing up a sequencing run

        # THEN the appropriate error should have been raised
        with pytest.raises(PdcError):
            backup_service.backup_run(
                files_to_archive=[
                    illumina_run_encryption_service.final_passphrase_file_path,
                    illumina_run_encryption_service.encrypted_gpg_file_path,
                ],
                store=base_store,
                sequencing_run=sequencing_run,
            )



import logging
from pathlib import Path

from cg.services.deliver_files.file_fetcher.models import CaseFile
from cg.services.deliver_files.file_formatter.files.abstract import FileFormatter
from cg.services.deliver_files.file_formatter.destination.models import FormattedFile
from cg.services.deliver_files.file_formatter.path_name.abstract import PathNameFormatter
from cg.services.deliver_files.file_formatter.path_name.nested_structure import (
    NestedStructurePathFormatter,
)
from cg.services.deliver_files.utils import FileManager

LOG = logging.getLogger(__name__)


class CaseFileFormatter(FileFormatter):
    """
    Format the case files to deliver and return the formatted files.
    args:
        path_name_formatter: The path name formatter to format paths to either a flat or nested structure in the delivery destination
        file_manager: The file manager
    """

    def __init__(
        self,
        path_name_formatter: PathNameFormatter,
        file_manager: FileManager,
    ):
        self.path_name_formatter = path_name_formatter
        self.file_manager = file_manager

    def format_files(self, moved_files: list[CaseFile], delivery_path: Path) -> list[FormattedFile]:
        """Format the case files to deliver and return the formatted files.
        args:
            moved_files: The case files to format
            delivery_path: The path to deliver the files to
        """
        LOG.debug("[FORMAT SERVICE] Formatting case files")
        self._create_case_name_folder(
            delivery_path=delivery_path, case_name=moved_files[0].case_name
        )
        return self._format_case_files(moved_files)

    def _format_case_files(self, case_files: list[CaseFile]) -> list[FormattedFile]:
        """Format the case files to deliver and return the formatted files.
        args:
            case_files: The case files to format
        """
        formatted_files: list[FormattedFile] = self._get_formatted_paths(case_files)
        for formatted_file in formatted_files:
            self.file_manager.rename_file(
                src=formatted_file.original_path, dst=formatted_file.formatted_path
            )
        return formatted_files

    def _create_case_name_folder(self, delivery_path: Path, case_name: str) -> None:
        """
        Create a folder for the case in the delivery path.
        The folder is only created if the provided PathStructureFormatter is a NestedStructurePathFormatter.
        args:
            delivery_path: The path to deliver the files to
            case_name: The name of the case
        """
        LOG.debug(f"[FORMAT SERVICE] Creating folder for case: {case_name}")
        if isinstance(self.path_name_formatter, NestedStructurePathFormatter):
            self.file_manager.create_directories(base_path=delivery_path, directories={case_name})

    def _get_formatted_paths(self, case_files: list[CaseFile]) -> list[FormattedFile]:
        """Return a list of formatted case files.
        args:
            case_files: The case files to format
        """
        formatted_files: list[FormattedFile] = []
        for case_file in case_files:
            formatted_path = self.path_name_formatter.format_file_path(
                file_path=case_file.file_path,
                provided_id=case_file.case_id,
                provided_name=case_file.case_name,
            )
            formatted_files.append(
                FormattedFile(original_path=case_file.file_path, formatted_path=formatted_path)
            )
        return formatted_files



"""Timestamp fixtures."""

from datetime import MAXYEAR, datetime, timedelta

import pytest


@pytest.fixture(scope="session")
def old_timestamp() -> datetime:
    """Return a time stamp in date time format."""
    return datetime(1900, 1, 1)


@pytest.fixture(scope="session")
def timestamp() -> datetime:
    """Return a time stamp in date time format."""
    return datetime(2020, 5, 1)


@pytest.fixture(scope="session")
def later_timestamp() -> datetime:
    """Return a time stamp in date time format."""
    return datetime(2020, 6, 1)


@pytest.fixture(scope="session")
def future_date() -> datetime:
    """Return a distant date in the future for which no events happen later."""
    return datetime(MAXYEAR, 1, 1, 1, 1, 1)


@pytest.fixture(scope="session")
def timestamp_now() -> datetime:
    """Return a time stamp of today's date in date time format."""
    return datetime.now()


@pytest.fixture(scope="session")
def timestamp_yesterday(timestamp_now: datetime) -> datetime:
    """Return a time stamp of yesterday's date in date time format."""
    return timestamp_now - timedelta(days=1)


@pytest.fixture(scope="session")
def timestamp_in_2_weeks(timestamp_now: datetime) -> datetime:
    """Return a time stamp 14 days ahead in time."""
    return timestamp_now + timedelta(days=14)



"""CLI support to create config and/or start NALLO."""

import logging

import rich_click as click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.constants.constants import MetaApis
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.nallo import NalloAnalysisAPI

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True, context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_context
def nallo(context: click.Context) -> None:
    """GMS/Nallo analysis workflow."""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis[MetaApis.ANALYSIS_API] = NalloAnalysisAPI(config=context.obj)



from pathlib import Path

from cg.meta.workflow.microsalt.metrics_parser import MetricsParser


def test_parse_valid_quality_metrics(microsalt_metrics_file: Path):
    # GIVEN a valid quality metrics file path

    # WHEN parsing the file
    MetricsParser.parse(microsalt_metrics_file)

    # THEN no error is thrown



import datetime as dt
from unittest.mock import Mock, patch

import pytest

from cg.clients.freshdesk.models import TicketResponse
from cg.constants import DataDelivery
from cg.constants.constants import Workflow
from cg.constants.subject import Sex
from cg.exc import OrderError, TicketCreationError
from cg.meta.orders import OrdersAPI
from cg.models.orders.order import OrderIn, OrderType
from cg.models.orders.samples import MipDnaSample
from cg.services.orders.validate_order_services.validate_case_order import (
    ValidateCaseOrderService,
)
from cg.store.models import Case, Customer, Pool, Sample, Order
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def monkeypatch_process_lims(monkeypatch, order_data) -> None:
    lims_project_data = {"id": "ADM1234", "date": dt.datetime.now()}
    lims_map = {sample.name: f"ELH123A{index}" for index, sample in enumerate(order_data.samples)}
    monkeypatch.setattr(
        "cg.services.orders.order_lims_service.order_lims_service.OrderLimsService.process_lims",
        lambda *args, **kwargs: (lims_project_data, lims_map),
    )


def mock_freshdesk_ticket_creation(mock_create_ticket, ticket_id: str):
    """Helper function to mock Freshdesk ticket creation."""
    mock_create_ticket.return_value = TicketResponse(
        id=int(ticket_id),
        description="This is a test description.",
        subject="Support needed..",
        status=2,
        priority=1,
    )


def mock_freshdesk_reply_to_ticket(mock_reply_to_ticket):
    """Helper function to mock Freshdesk reply to ticket."""
    mock_reply_to_ticket.return_value = None


def test_too_long_order_name():
    # GIVEN order with more than allowed characters name
    long_name = "A super long order name that is longer than sixty-four characters."
    assert len(long_name) > Sample.order.property.columns[0].type.length

    # WHEN placing it in the pydantic order model
    # THEN an error is raised
    with pytest.raises(ValueError):
        OrderIn(name=long_name, customer="", comment="", samples=[])


@pytest.mark.parametrize(
    "order_type",
    [
        OrderType.BALSAMIC,
        OrderType.FASTQ,
        OrderType.FLUFFY,
        OrderType.METAGENOME,
        OrderType.MICROSALT,
        OrderType.MIP_DNA,
        OrderType.MIP_RNA,
        OrderType.RML,
        OrderType.RNAFUSION,
        OrderType.SARS_COV_2,
    ],
)
def test_submit(
    all_orders_to_submit: dict,
    base_store: Store,
    monkeypatch: pytest.MonkeyPatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    ticket_id: str,
    user_mail: str,
    user_name: str,
):
    with patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.create_ticket"
    ) as mock_create_ticket, patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.reply_to_ticket"
    ) as mock_reply_to_ticket:
        mock_freshdesk_ticket_creation(mock_create_ticket, ticket_id)
        mock_freshdesk_reply_to_ticket(mock_reply_to_ticket)

        order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
        monkeypatch_process_lims(monkeypatch, order_data)

        # GIVEN an order and an empty store
        assert not base_store._get_query(table=Sample).first()

        # WHEN submitting the order

        result = orders_api.submit(
            project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
        )

        # THEN the result should contain the ticket number for the order
        for record in result["records"]:
            if isinstance(record, Pool):
                assert record.ticket == ticket_id
            elif isinstance(record, Sample):
                assert record.original_ticket == ticket_id
            elif isinstance(record, Case):
                for link_obj in record.links:
                    assert link_obj.sample.original_ticket == ticket_id


@pytest.mark.parametrize(
    "order_type",
    [OrderType.MIP_DNA, OrderType.MIP_RNA, OrderType.BALSAMIC],
)
def test_submit_ticketexception(
    all_orders_to_submit,
    orders_api: OrdersAPI,
    order_type: OrderType,
    user_mail: str,
    user_name: str,
):
    # GIVEN a mock Freshdesk ticket creation that raises TicketCreationError
    with patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.create_ticket",
        side_effect=TicketCreationError("ERROR"),
    ):
        # GIVEN an order that does not have a name (ticket_nr)
        order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
        order_data.name = "dummy_name"

        # WHEN the order is submitted and a TicketCreationError raised
        # THEN the TicketCreationError is not excepted
        with pytest.raises(TicketCreationError):
            orders_api.submit(
                project=order_type,
                order_in=order_data,
                user_name=user_name,
                user_mail=user_mail,
            )


@pytest.mark.parametrize(
    "order_type",
    [OrderType.MIP_DNA, OrderType.MIP_RNA, OrderType.BALSAMIC],
)
def test_submit_illegal_sample_customer(
    all_orders_to_submit: dict,
    monkeypatch: pytest.MonkeyPatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    sample_store: Store,
    user_mail: str,
    user_name: str,
):
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    monkeypatch_process_lims(monkeypatch, order_data)
    # GIVEN we have an order with a customer that is not in the same customer group as customer
    # that the samples originate from
    new_customer = sample_store.add_customer(
        "customer999",
        "customer 999",
        scout_access=True,
        invoice_address="dummy street",
        invoice_reference="dummy nr",
    )
    sample_store.session.add(new_customer)
    existing_sample: Sample = sample_store._get_query(table=Sample).first()
    existing_sample.customer = new_customer
    sample_store.session.add(existing_sample)
    sample_store.session.commit()
    for sample in order_data.samples:
        sample.internal_id = existing_sample.internal_id

    # WHEN calling submit
    # THEN an OrderError should be raised on illegal customer
    with pytest.raises(OrderError):
        orders_api.submit(
            project=order_type,
            order_in=order_data,
            user_name=user_name,
            user_mail=user_mail,
        )


@pytest.mark.parametrize(
    "order_type",
    [OrderType.MIP_DNA, OrderType.MIP_RNA, OrderType.BALSAMIC],
)
def test_submit_scout_legal_sample_customer(
    all_orders_to_submit: dict,
    monkeypatch: pytest.MonkeyPatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    sample_store: Store,
    user_mail: str,
    user_name: str,
    ticket_id: str,
):
    with patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.create_ticket"
    ) as mock_create_ticket, patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.reply_to_ticket"
    ) as mock_reply_to_ticket:
        mock_freshdesk_ticket_creation(mock_create_ticket, ticket_id)
        mock_freshdesk_reply_to_ticket(mock_reply_to_ticket)
        order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
        monkeypatch_process_lims(monkeypatch, order_data)
        # GIVEN we have an order with a customer that is in the same customer group as customer
        # that the samples originate from
        collaboration = sample_store.add_collaboration("customer999only", "customer 999 only group")
        sample_store.session.add(collaboration)
        sample_customer = sample_store.add_customer(
            "customer1",
            "customer 1",
            scout_access=True,
            invoice_address="dummy street 1",
            invoice_reference="dummy nr",
        )
        order_customer = sample_store.add_customer(
            "customer2",
            "customer 2",
            scout_access=True,
            invoice_address="dummy street 2",
            invoice_reference="dummy nr",
        )
        sample_customer.collaborations.append(collaboration)
        order_customer.collaborations.append(collaboration)
        sample_store.session.add(sample_customer)
        sample_store.session.add(order_customer)
        existing_sample: Sample = sample_store._get_query(table=Sample).first()
        existing_sample.customer = sample_customer
        sample_store.session.commit()
        order_data.customer = order_customer.internal_id

        for sample in order_data.samples:
            sample.internal_id = existing_sample.internal_id
            break

        # WHEN calling submit
        # THEN an OrderError should not be raised on illegal customer
        orders_api.submit(
            project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
        )


@pytest.mark.parametrize(
    "order_type",
    [OrderType.MIP_DNA, OrderType.MIP_RNA, OrderType.BALSAMIC],
)
def test_submit_duplicate_sample_case_name(
    all_orders_to_submit: dict,
    monkeypatch: pytest.MonkeyPatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    ticket_id: str,
    user_mail: str,
    user_name: str,
):
    # GIVEN we have an order with a case that is already in the database
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    store = orders_api.status
    customer: Customer = store.get_customer_by_internal_id(customer_internal_id=order_data.customer)
    for sample in order_data.samples:
        case_id = sample.family_name
        if not store.get_case_by_name_and_customer(customer=customer, case_name=case_id):
            case: Case = store.add_case(
                data_analysis=Workflow.MIP_DNA,
                data_delivery=DataDelivery.SCOUT,
                name=case_id,
                ticket=ticket_id,
            )
            case.customer = customer
            store.session.add(case)
        store.session.commit()
        assert store.get_case_by_name_and_customer(customer=customer, case_name=case_id)

    monkeypatch_process_lims(monkeypatch, order_data)

    # WHEN calling submit
    # THEN an OrderError should be raised on duplicate case name
    with pytest.raises(OrderError):
        orders_api.submit(
            project=order_type,
            order_in=order_data,
            user_name=user_name,
            user_mail=user_mail,
        )


@pytest.mark.parametrize(
    "order_type",
    [OrderType.FLUFFY],
)
def test_submit_fluffy_duplicate_sample_case_name(
    all_orders_to_submit: dict,
    monkeypatch: pytest.MonkeyPatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    user_mail: str,
    user_name: str,
    ticket_id: str,
):
    with patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.create_ticket"
    ) as mock_create_ticket, patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.reply_to_ticket"
    ) as mock_reply_to_ticket:
        mock_freshdesk_ticket_creation(mock_create_ticket, ticket_id)
        mock_freshdesk_reply_to_ticket(mock_reply_to_ticket)
        # GIVEN we have an order with a case that is already in the database
        order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
        monkeypatch_process_lims(monkeypatch, order_data)

        orders_api.submit(
            project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
        )

        # WHEN calling submit
        # THEN an OrderError should be raised on duplicate case name
        with pytest.raises(OrderError):
            orders_api.submit(
                project=order_type,
                order_in=order_data,
                user_name=user_name,
                user_mail=user_mail,
            )


def test_submit_unique_sample_case_name(
    orders_api: OrdersAPI,
    mip_order_to_submit: dict,
    user_name: str,
    user_mail: str,
    monkeypatch: pytest.MonkeyPatch,
    ticket_id: str,
):
    with patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.create_ticket"
    ) as mock_create_ticket, patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.reply_to_ticket"
    ) as mock_reply_to_ticket:
        mock_freshdesk_ticket_creation(mock_create_ticket, ticket_id)
        mock_freshdesk_reply_to_ticket(mock_reply_to_ticket)

        # GIVEN we have an order with a case that is not existing in the database
        order_data = OrderIn.parse_obj(obj=mip_order_to_submit, project=OrderType.MIP_DNA)

        store = orders_api.status

        sample: MipDnaSample
        for sample in order_data.samples:
            case_id = sample.family_name
            customer: Customer = store.get_customer_by_internal_id(
                customer_internal_id=order_data.customer
            )
            assert not store.get_case_by_name_and_customer(customer=customer, case_name=case_id)

        monkeypatch_process_lims(monkeypatch, order_data)

        # WHEN calling submit
        orders_api.submit(
            project=OrderType.MIP_DNA,
            order_in=order_data,
            user_name=user_name,
            user_mail=user_mail,
        )

        # Then no exception about duplicate names should be thrown


def test_validate_sex_inconsistent_sex(
    orders_api: OrdersAPI, mip_order_to_submit: dict, helpers: StoreHelpers
):
    # GIVEN we have an order with a sample that is already in the database but with different sex
    order_data = OrderIn.parse_obj(mip_order_to_submit, project=OrderType.MIP_DNA)
    store = orders_api.status
    customer: Customer = store.get_customer_by_internal_id(customer_internal_id=order_data.customer)

    # add sample with different sex than in order
    sample: MipDnaSample
    for sample in order_data.samples:
        sample_obj: Sample = helpers.add_sample(
            store=store,
            customer_id=customer.internal_id,
            sex=Sex.MALE if sample.sex == Sex.FEMALE else Sex.FEMALE,
            name=sample.name,
            subject_id=sample.subject_id,
        )
        store.session.add(sample_obj)
        store.session.commit()
        assert sample_obj.sex != sample.sex

    validator = ValidateCaseOrderService(status_db=orders_api.status)

    # WHEN calling _validate_sex
    # THEN an OrderError should be raised on non-matching sex
    with pytest.raises(OrderError):
        validator._validate_subject_sex(samples=order_data.samples, customer_id=order_data.customer)


def test_validate_sex_consistent_sex(
    orders_api: OrdersAPI, mip_order_to_submit: dict, helpers: StoreHelpers
):
    # GIVEN we have an order with a sample that is already in the database and with same gender
    order_data = OrderIn.parse_obj(mip_order_to_submit, project=OrderType.MIP_DNA)
    store = orders_api.status
    customer: Customer = store.get_customer_by_internal_id(customer_internal_id=order_data.customer)

    # add sample with different sex than in order
    sample: MipDnaSample
    for sample in order_data.samples:
        sample_obj: Sample = helpers.add_sample(
            store=store,
            customer_id=customer.internal_id,
            sex=sample.sex,
            name=sample.name,
            subject_id=sample.subject_id,
        )
        store.session.add(sample_obj)
        store.session.commit()
        assert sample_obj.sex == sample.sex

    validator = ValidateCaseOrderService(status_db=orders_api.status)

    # WHEN calling _validate_sex
    validator._validate_subject_sex(samples=order_data.samples, customer_id=order_data.customer)

    # THEN no OrderError should be raised on non-matching sex


def test_validate_sex_unknown_existing_sex(
    orders_api: OrdersAPI, mip_order_to_submit: dict, helpers: StoreHelpers
):
    # GIVEN we have an order with a sample that is already in the database and with different gender but the existing is
    # of type "unknown"
    order_data = OrderIn.parse_obj(mip_order_to_submit, project=OrderType.MIP_DNA)
    store = orders_api.status
    customer: Customer = store.get_customer_by_internal_id(customer_internal_id=order_data.customer)

    # add sample with different sex than in order
    sample: MipDnaSample
    for sample in order_data.samples:
        sample_obj: Sample = helpers.add_sample(
            store=store,
            customer_id=customer.internal_id,
            sex=Sex.UNKNOWN,
            name=sample.name,
            subject_id=sample.subject_id,
        )
        store.session.add(sample_obj)
        store.session.commit()
        assert sample_obj.sex != sample.sex

    validator = ValidateCaseOrderService(status_db=orders_api.status)

    # WHEN calling _validate_sex
    validator._validate_subject_sex(samples=order_data.samples, customer_id=order_data.customer)

    # THEN no OrderError should be raised on non-matching sex


def test_validate_sex_unknown_new_sex(
    orders_api: OrdersAPI, mip_order_to_submit: dict, helpers: StoreHelpers
):
    # GIVEN we have an order with a sample that is already in the database and with different gender but the new is of
    # type "unknown"
    order_data = OrderIn.parse_obj(mip_order_to_submit, project=OrderType.MIP_DNA)
    store = orders_api.status
    customer: Customer = store.get_customer_by_internal_id(customer_internal_id=order_data.customer)

    # add sample with different sex than in order
    for sample in order_data.samples:
        sample_obj: Sample = helpers.add_sample(
            store=store,
            customer_id=customer.internal_id,
            sex=sample.sex,
            name=sample.name,
            subject_id=sample.subject_id,
        )
        sample.sex = "unknown"
        store.session.add(sample_obj)
        store.session.commit()

    for sample in order_data.samples:
        assert sample_obj.sex != sample.sex

    validator = ValidateCaseOrderService(status_db=orders_api.status)

    # WHEN calling _validate_sex
    validator._validate_subject_sex(samples=order_data.samples, customer_id=order_data.customer)

    # THEN no OrderError should be raised on non-matching sex


@pytest.mark.parametrize(
    "order_type",
    [
        OrderType.BALSAMIC,
        OrderType.FASTQ,
        OrderType.FLUFFY,
        OrderType.METAGENOME,
        OrderType.MICROSALT,
        OrderType.MIP_DNA,
        OrderType.MIP_RNA,
        OrderType.RML,
        OrderType.SARS_COV_2,
    ],
)
def test_submit_unique_sample_name(
    all_orders_to_submit: dict,
    monkeypatch: pytest.MonkeyPatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    user_mail: str,
    user_name: str,
    ticket_id: str,
):
    with patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.create_ticket"
    ) as mock_create_ticket, patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.reply_to_ticket"
    ) as mock_reply_to_ticket:
        mock_freshdesk_ticket_creation(mock_create_ticket, ticket_id)
        mock_freshdesk_reply_to_ticket(mock_reply_to_ticket)
        # GIVEN we have an order with a sample that is not existing in the database
        order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
        store = orders_api.status
        assert not store._get_query(table=Sample).first()

        monkeypatch_process_lims(monkeypatch, order_data)

        # WHEN calling submit
        orders_api.submit(
            project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
        )

        # Then no exception about duplicate names should be thrown


@pytest.mark.parametrize(
    "order_type",
    [OrderType.SARS_COV_2, OrderType.METAGENOME],
)
def test_sarscov2_submit_duplicate_sample_name(
    all_orders_to_submit: dict,
    helpers: StoreHelpers,
    monkeypatch: pytest.MonkeyPatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    user_mail: str,
    user_name: str,
):
    # GIVEN we have an order with samples that is already in the database
    order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
    monkeypatch_process_lims(monkeypatch, order_data)
    store_samples_with_names_from_order(orders_api.status, helpers, order_data)

    # WHEN calling submit
    # THEN an OrderError should be raised on duplicate sample name
    with pytest.raises(OrderError):
        orders_api.submit(
            project=order_type,
            order_in=order_data,
            user_name=user_name,
            user_mail=user_mail,
        )


def store_samples_with_names_from_order(store: Store, helpers: StoreHelpers, order_data: OrderIn):
    customer: Customer = store.get_customer_by_internal_id(customer_internal_id=order_data.customer)
    for sample in order_data.samples:
        sample_name = sample.name
        if not store.get_sample_by_customer_and_name(
            customer_entry_id=[customer.id], sample_name=sample_name
        ):
            sample_obj = helpers.add_sample(
                store=store, customer_id=customer.internal_id, name=sample_name
            )
            store.session.add(sample_obj)
            store.session.commit()


@pytest.mark.parametrize(
    "order_type",
    [
        OrderType.BALSAMIC,
        OrderType.FASTQ,
        OrderType.MICROSALT,
        OrderType.MIP_DNA,
        OrderType.MIP_RNA,
        OrderType.RML,
    ],
)
def test_not_sarscov2_submit_duplicate_sample_name(
    all_orders_to_submit: dict,
    helpers: StoreHelpers,
    monkeypatch: pytest.MonkeyPatch,
    order_type: OrderType,
    orders_api: OrdersAPI,
    user_mail: str,
    user_name: str,
    ticket_id: str,
):
    with patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.create_ticket"
    ) as mock_create_ticket, patch(
        "cg.clients.freshdesk.freshdesk_client.FreshdeskClient.reply_to_ticket"
    ) as mock_reply_to_ticket:
        mock_freshdesk_ticket_creation(mock_create_ticket, ticket_id)
        mock_freshdesk_reply_to_ticket(mock_reply_to_ticket)
        # GIVEN we have an order with samples that is already in the database
        order_data = OrderIn.parse_obj(obj=all_orders_to_submit[order_type], project=order_type)
        monkeypatch_process_lims(monkeypatch, order_data)
        store_samples_with_names_from_order(orders_api.status, helpers, order_data)

        # WHEN calling submit
        orders_api.submit(
            project=order_type, order_in=order_data, user_name=user_name, user_mail=user_mail
        )

        # THEN no OrderError should be raised on duplicate sample name



"""Delivery API related models."""

from pathlib import Path

from pydantic import BaseModel, FilePath


class DeliveryFile(BaseModel):
    """Model representing a file to be delivered."""

    source_path: FilePath
    destination_path: Path



from cg.constants import Workflow, SequencingFileTag
from cg.services.deliver_files.tag_fetcher.abstract import FetchDeliveryFileTagsService
from cg.services.deliver_files.tag_fetcher.error_handling import handle_tag_errors
from cg.services.deliver_files.tag_fetcher.models import DeliveryFileTags


class FOHMUploadTagsFetcher(FetchDeliveryFileTagsService):
    """Class to fetch tags for FOHM upload files."""

    @handle_tag_errors
    def fetch_tags(self, workflow: Workflow) -> DeliveryFileTags:
        """
        Fetch the tags for the bam files to deliver.
        NOTE: workflow raw data here is required to fit the implementation of the raw data delivery file fetcher.
        if workflow is MUTANT, return tags for consensus-sample and vcf-report to fetch sample files from the case bundle.
        if workflow is RAW_DATA, return tags for fastq to fetch fastq files from the sample bundle.
        Required since some of the sample specific files are stored on the case bundle, but also fastq files.
        Not separating these would cause fetching of case bundle fastq files if present.

        Hardcoded to only return the tags for the files to deliver.
        args:
            workflow: Workflow: The workflow to fetch tags
        """
        self._validate_workflow(workflow=workflow)
        return (
            DeliveryFileTags(
                case_tags=None,
                sample_tags=[{"consensus-sample"}, {"vcf-report"}],
            )
            if workflow == Workflow.MUTANT
            else DeliveryFileTags(
                case_tags=None,
                sample_tags=[{SequencingFileTag.FASTQ}],
            )
        )

    @staticmethod
    def _validate_workflow(workflow: Workflow):
        """
        Validate the workflow.
        NOTE: workflow raw data here is required to fit the implementation of the raw data delivery file fetcher.
        args:
            workflow: Workflow: The workflow to validate.
        """
        if workflow not in [Workflow.MUTANT, Workflow.RAW_DATA]:
            raise ValueError(f"Workflow {workflow} is not supported for FOHM upload file delivery.")
        return workflow



import pytest

from cg.services.deliver_files.file_fetcher.abstract import (
    FetchDeliveryFilesService,
)
from cg.services.deliver_files.file_fetcher.exc import NoDeliveryFilesError
from cg.services.deliver_files.file_fetcher.models import DeliveryFiles


@pytest.mark.parametrize(
    "expected_delivery_files,delivery_file_service,sample_id_to_fetch",
    [
        ("expected_fohm_delivery_files", "fohm_data_delivery_service", "empty_sample"),
        ("expected_fastq_delivery_files", "raw_data_delivery_service", "empty_sample"),
        ("expected_analysis_delivery_files", "analysis_delivery_service", "empty_sample"),
        ("expected_bam_delivery_files", "bam_data_delivery_service", "empty_sample"),
        ("expected_bam_delivery_files_single_sample", "bam_data_delivery_service", "sample_id"),
    ],
)
def test_get_files_to_deliver(
    expected_delivery_files: DeliveryFiles,
    delivery_file_service: FetchDeliveryFilesService,
    sample_id_to_fetch: str | None,
    case_id: str,
    request,
):
    """Test to get the files to deliver from the FetchDeliveryFilesService."""
    # GIVEN a case id, samples that are present in Housekeeper and a delivery service
    delivery_file_service = request.getfixturevalue(delivery_file_service)
    expected_delivery_files = request.getfixturevalue(expected_delivery_files)
    sample_id: str | None = request.getfixturevalue(sample_id_to_fetch)

    # WHEN getting the files to deliver
    delivery_files: DeliveryFiles = delivery_file_service.get_files_to_deliver(
        case_id=case_id, sample_id=sample_id
    )

    # THEN assert that the files to deliver are fetched
    assert delivery_files == expected_delivery_files


@pytest.mark.parametrize(
    "delivery_file_service",
    [
        "raw_data_delivery_service_no_housekeeper_bundle",
        "analysis_delivery_service_no_housekeeper_bundle",
        "bam_data_delivery_service_no_housekeeper_bundle",
    ],
)
def test_get_files_to_deliver_no_files_present(
    delivery_file_service: FetchDeliveryFilesService,
    case_id: str,
    request,
):
    """Test to get the files to deliver from the FetchDeliveryFilesService when no files are present."""

    # GIVEN a case id, samples that are not present in housekeeper and a delivery service
    delivery_file_service = request.getfixturevalue(delivery_file_service)

    # WHEN getting the files to deliver that raises an NoDeliveryFilesError

    # THEN assert that the NoDeliveryFilesError is raised
    with pytest.raises(NoDeliveryFilesError):
        delivery_file_service.get_files_to_deliver(case_id)



import pytest

from cg.exc import OrderError
from cg.models.orders.order import OrderIn
from cg.services.orders.validate_order_services.validate_pacbio_order import (
    ValidatePacbioOrderService,
)
from cg.store.store import Store


def test_validate_valid_pacbio_order(
    validate_pacbio_order_service: ValidatePacbioOrderService, pacbio_order: OrderIn
):
    # GIVEN a valid PacBio order

    # WHEN validating the order
    validate_pacbio_order_service.validate_order(pacbio_order)

    # THEN no error is raised


def test_validate_pacbio_order_unknown_customer(
    pacbio_order: OrderIn, validate_pacbio_order_service: ValidatePacbioOrderService
):
    # GIVEN a PacBio order with an unknown customer
    pacbio_order.customer = "unknown_customer"

    # WHEN validating the order

    # THEN an order error should be raised
    with pytest.raises(OrderError):
        validate_pacbio_order_service.validate_order(pacbio_order)


def test_validate_pacbio_order_invalid_application(
    pacbio_order: OrderIn, validate_pacbio_order_service: ValidatePacbioOrderService
):
    # GIVEN a PacBio order with an unknown application
    pacbio_order.samples[0].application = "unknown_application"

    # WHEN validating the order

    # THEN an order error should be raised
    with pytest.raises(OrderError):
        validate_pacbio_order_service.validate_order(pacbio_order)


def test_validate_pacbio_order_reused_sample_name(
    pacbio_order: OrderIn, validate_pacbio_order_service: ValidatePacbioOrderService
):
    # GIVEN a PacBio order with a reused sample name
    status_db: Store = validate_pacbio_order_service.status_db
    customer = status_db.get_customer_by_internal_id(pacbio_order.customer)
    old_sample_name: str = status_db.get_samples_by_customers_and_pattern(customers=[customer])[
        0
    ].name
    pacbio_order.samples[0].name = old_sample_name

    # WHEN validating the order

    # THEN an order error should be raised
    with pytest.raises(OrderError):
        validate_pacbio_order_service.validate_order(pacbio_order)



"""Test MIP config file"""

from pathlib import Path

from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile
from cg.models.mip.mip_config import AnalysisType, MipBaseConfig


def test_instantiate_mip_config(mip_analysis_config_dna_raw: dict):
    """
    Tests mip config against a pydantic MipBaseConfig
    """
    # GIVEN a dictionary with the basic config

    # WHEN instantiating a MipBaseConfig object
    config_object = MipBaseConfig(**mip_analysis_config_dna_raw)

    # THEN assert that it was successfully created
    assert isinstance(config_object, MipBaseConfig)


def test_mip_config(mip_case_config_dna: Path):
    """Test to parse the content of a real MIP DNA config file"""
    # GIVEN the path to a file with config metadata content
    raw_config: dict = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=mip_case_config_dna
    )

    # WHEN instantiating a MipBaseSampleInfo object
    config_object = MipBaseConfig(**raw_config)

    # THEN assert that it was successfully created
    assert isinstance(config_object, MipBaseConfig)


def test_mip_config_case_id(mip_analysis_config_dna_raw: dict):
    """Test case_id validator"""
    # GIVEN a MIP config

    # WHEN instantiating a MipBaseSampleInfo object
    config_object = MipBaseConfig(**mip_analysis_config_dna_raw)

    # THEN assert that case_id was set
    assert config_object.case_id == "yellowhog"


def test_mip_config_case_id_with_family_id(mip_analysis_config_dna_raw: dict):
    """Test case_id validator"""
    # GIVEN a MIP config missing a case_id but with family_id
    mip_analysis_config_dna_raw.pop("case_id")

    # WHEN instantiating a MipBaseSampleInfo object
    config_object = MipBaseConfig(**mip_analysis_config_dna_raw)

    # THEN assert that case_id was set
    assert config_object.case_id == "a_family_id"


def test_mip_config_analysis_type(mip_analysis_config_dna_raw: dict):
    """Test analysis type validator."""
    # GIVEN a MIP config

    # WHEN instantiating a MipBaseSampleInfo object
    config_object = MipBaseConfig(**mip_analysis_config_dna_raw)

    # THEN assert that samples were set
    assert config_object.samples

    # THEN assert that dict is set
    analysis_type: AnalysisType = config_object.samples.pop()
    assert analysis_type.model_dump() == {"analysis_type": "wgs", "sample_id": "sample_id"}



from pathlib import Path

from cg.constants.constants import FileFormat
from cg.models.nf_analysis import FileDeliverable


def test_file_deliverables(any_string: str, filled_file: Path):
    """Tests that file deliverable model is successfully created."""
    # GIVEN valid deliverables fields

    # WHEN instantiating a deliverable object
    file_deliverable = FileDeliverable(
        format=FileFormat.TSV,
        id=any_string,
        path=filled_file,
        path_index=filled_file,
        step=any_string,
        tag=any_string,
    )

    # THEN assert that it was successfully created
    assert isinstance(file_deliverable, FileDeliverable)

    # THEN assert that paths are returned as strings
    assert isinstance(file_deliverable.path, str)
    assert isinstance(file_deliverable.path_index, str)


def test_file_deliverables_missing_optional(any_string: str, filled_file: Path):
    """Tests file delivery when an optional field is missing."""
    # GIVEN valid deliverables fields

    # WHEN instantiating a deliverable object
    file_deliverable = FileDeliverable(
        format=FileFormat.TSV,
        id=any_string,
        path=filled_file,
        step=any_string,
        tag=any_string,
    )

    # THEN assert that it was successfully created
    assert isinstance(file_deliverable, FileDeliverable)


def test_file_deliverables_non_existing_attribute(any_string: str, filled_file: Path):
    """Tests file delivery when a non-existing attribute is given."""
    # WHEN instantiating a deliverable object with additional attributes not present in the model
    file_deliverable = FileDeliverable(
        format=FileFormat.TSV,
        id=any_string,
        path=filled_file,
        step=any_string,
        tag=any_string,
        nonexisting=FileFormat.TSV,
    )

    # THEN assert that it was successfully created
    assert isinstance(file_deliverable, FileDeliverable)

    # THEN assert that not existing attribute is not included
    assert "nonexisting" not in file_deliverable.__annotations__.keys()


def test_file_deliverables_non_existing_file(
    any_string: str,
    non_existing_file_path: Path,
):
    """Tests file delivery when a mandatory file does not exist."""
    # WHEN instantiating a deliverables object

    # THEN assert that an error is raised
    file_deliverable = FileDeliverable(
        format=FileFormat.TSV,
        id=any_string,
        path=non_existing_file_path,
        step=any_string,
        tag=any_string,
    )
    # THEN assert that it was successfully created
    assert isinstance(file_deliverable, FileDeliverable)



"""This script tests the cli methods to set families to status-db"""

import logging

from click.testing import CliRunner

from cg.cli.delete.case import delete_case as delete_case_command
from cg.models.cg_config import CGConfig
from cg.store.models import Case, CaseSample, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers

SUCCESS = 0


def test_delete_case_without_options(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test to delete a case using only the required arguments"""
    # GIVEN a database with a case
    base_store: Store = base_context.status_db
    helpers.add_case(base_store)
    case_query = base_store._get_query(table=Case)
    assert case_query.count() == 1

    # WHEN deleting a case
    result = cli_runner.invoke(delete_case_command, obj=base_context)

    # THEN it should abort
    assert result.exit_code != SUCCESS


def test_delete_case_bad_case(cli_runner: CliRunner, base_context: CGConfig):
    """Test to delete a case using a non-existing case"""
    # GIVEN an empty database

    # WHEN deleting a case
    case_id = "dummy_name"
    result = cli_runner.invoke(delete_case_command, [case_id], obj=base_context)

    # THEN it should complain on missing case
    assert result.exit_code != SUCCESS


def test_delete_case_without_links(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test that the delete case can delete a case without links"""
    # GIVEN a database with a case
    base_store: Store = base_context.status_db
    case_obj = helpers.add_case(base_store)
    case_id = case_obj.internal_id
    assert not case_obj.links

    # WHEN deleting a case
    result = cli_runner.invoke(delete_case_command, [case_id, "--yes"], obj=base_context)
    # THEN it should have been deleted
    assert result.exit_code == SUCCESS
    assert base_store._get_query(table=Case).count() == 0


def test_delete_case_with_analysis(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test that the delete case can't delete a case with analysis"""
    # GIVEN a database with a case with an analysis
    base_store: Store = base_context.status_db
    analysis_obj = helpers.add_analysis(base_store)
    case_id = analysis_obj.case.internal_id

    # WHEN deleting a case

    result = cli_runner.invoke(delete_case_command, [case_id, "--yes"], obj=base_context)

    # THEN it should not have been deleted
    assert result.exit_code != SUCCESS
    assert base_store._get_query(table=Case).count() == 1


def test_delete_case_with_dry_run(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers, caplog
):
    """Test that the delete case will not delete the case in dry-run mode"""
    # GIVEN a database with a case
    base_store: Store = base_context.status_db
    case_obj = helpers.add_case(base_store)
    case_id = case_obj.internal_id
    sample = helpers.add_sample(base_store)
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample)

    case_query = base_store._get_query(table=Case)
    family_sample_query = base_store._get_query(table=CaseSample)
    sample_query = base_store._get_query(table=Sample)

    assert case_query.count() == 1
    assert family_sample_query.count() == 1
    assert sample_query.count() == 1

    # WHEN deleting a case
    caplog.set_level(logging.DEBUG)
    result = cli_runner.invoke(
        delete_case_command, [case_id, "--yes", "--dry-run"], obj=base_context
    )

    # THEN it should not have been deleted
    assert result.exit_code == SUCCESS
    assert case_query.count() == 1
    assert family_sample_query.count() == 1
    assert sample_query.count() == 1
    assert "Link:" in caplog.text
    assert "Sample is linked" in caplog.text
    assert "Case:" in caplog.text
    assert " was NOT deleted due to --dry-run" in caplog.text


def test_delete_case_without_yes(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test that the delete case will not delete the case in dry-run mode"""
    # GIVEN a database with a case
    base_store: Store = base_context.status_db
    case_obj = helpers.add_case(base_store)
    case_id = case_obj.internal_id
    assert not case_obj.links

    # WHEN deleting a case
    result = cli_runner.invoke(delete_case_command, [case_id], obj=base_context)

    # THEN it should not have been deleted
    assert result.exit_code != SUCCESS
    case_query = base_store._get_query(table=Case)
    assert case_query.count() == 1


def test_delete_case_with_links(cli_runner: CliRunner, base_context: CGConfig, helpers):
    """Test that the delete case can delete a case without links"""
    # GIVEN a database with a case
    base_store: Store = base_context.status_db
    case_obj = helpers.add_case(base_store)
    case_id = case_obj.internal_id
    sample = helpers.add_sample(base_store)
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample)

    case_query = base_store._get_query(table=Case)
    family_sample_query = base_store._get_query(table=CaseSample)
    sample_query = base_store._get_query(table=Sample)

    assert case_query.count() > 0
    assert family_sample_query.count() > 0
    assert sample_query.count() > 0

    # WHEN deleting a case with links
    result = cli_runner.invoke(delete_case_command, [case_id, "--yes"], obj=base_context)

    # THEN it should have been deleted
    assert result.exit_code == SUCCESS
    assert case_query.count() == 0
    assert family_sample_query.count() == 0
    assert sample_query.count() == 0


def test_delete_case_with_links_to_other_case(
    cli_runner: CliRunner, base_context: CGConfig, helpers
):
    """Test that the delete case will not delete a sample linked to another case"""
    # GIVEN a database with a case
    base_store: Store = base_context.status_db
    case_obj = helpers.add_case(base_store, "first_case_linked_to_sample")
    case_id = case_obj.internal_id
    sample = helpers.add_sample(base_store)
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample)
    case_obj2 = helpers.add_case(base_store, "second_case_linked_to_sample")
    helpers.add_relationship(store=base_store, case=case_obj2, sample=sample)

    case_query = base_store._get_query(table=Case)
    family_sample_query = base_store._get_query(table=CaseSample)
    sample_query = base_store._get_query(table=Sample)

    assert case_query.count() == 2
    assert family_sample_query.count() == 2
    assert sample_query.count() == 1

    # WHEN deleting a case
    result = cli_runner.invoke(delete_case_command, [case_id, "--yes"], obj=base_context)

    # THEN the first case should be gone with its link to the sample but not the other or
    # the sample
    assert result.exit_code == SUCCESS
    assert case_query.count() == 1
    assert family_sample_query.count() == 1
    assert sample_query.count() == 1


def test_delete_case_with_father_links(
    cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers
):
    """Test that the delete case will not delete a sample linked to another case as father"""
    # GIVEN a database with a case
    base_store: Store = base_context.status_db
    case_obj = helpers.add_case(base_store, "first_case_linked_to_sample")
    case_id = case_obj.internal_id
    sample_father = helpers.add_sample(base_store, "father")
    sample_child = helpers.add_sample(base_store, "child")
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample_father)
    case_obj2 = helpers.add_case(base_store, "second_case_linked_to_sample")
    helpers.add_relationship(
        store=base_store, case=case_obj2, sample=sample_child, father=sample_father
    )
    case_query = base_store._get_query(table=Case)
    family_sample_query = base_store._get_query(table=CaseSample)
    sample_query = base_store._get_query(table=Sample)

    assert case_query.count() == 2
    assert family_sample_query.count() == 2
    assert sample_query.count() == 2

    # WHEN deleting a case
    result = cli_runner.invoke(delete_case_command, [case_id, "--yes"], obj=base_context)

    # THEN the first case should be gone with its link to the sample but not the other or
    # the father sample
    assert result.exit_code == SUCCESS
    assert case_query.count() == 1
    assert family_sample_query.count() == 1
    assert sample_query.count() == 2


def test_delete_mother_case(cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers):
    """Test that the delete case will not delete a sample linked to another case as mother"""
    # GIVEN a database with a mother case and a child case with the mother as mother
    base_store: Store = base_context.status_db
    case_mother = helpers.add_case(base_store, "case_mother")
    case_mother_id = case_mother.internal_id
    sample_mother = helpers.add_sample(base_store, "mother")
    sample_child = helpers.add_sample(base_store, "child")
    helpers.add_relationship(store=base_store, case=case_mother, sample=sample_mother)
    case_child = helpers.add_case(base_store, "case_child")
    helpers.add_relationship(
        store=base_store, case=case_child, sample=sample_child, mother=sample_mother
    )

    case_query = base_store._get_query(table=Case)
    family_sample_query = base_store._get_query(table=CaseSample)
    sample_query = base_store._get_query(table=Sample)

    assert case_query.count() == 2
    assert family_sample_query.count() == 2
    assert sample_query.count() == 2

    # WHEN deleting the mother case
    result = cli_runner.invoke(delete_case_command, [case_mother_id, "--yes"], obj=base_context)

    # THEN the mother sample is not deletable
    assert result.exit_code == SUCCESS
    assert case_query.count() == 1
    assert family_sample_query.count() == 1
    assert sample_query.count() == 2


def test_delete_child_case(cli_runner: CliRunner, base_context: CGConfig, helpers: StoreHelpers):
    """Test that the delete case will not delete a sample linked to another case as mother"""
    # GIVEN a database with a mother case and a child case with the mother as mother
    base_store: Store = base_context.status_db
    case_mother = helpers.add_case(base_store, "case_mother")
    sample_mother = helpers.add_sample(base_store, "mother")
    sample_child = helpers.add_sample(base_store, "child")
    helpers.add_relationship(store=base_store, case=case_mother, sample=sample_mother)
    case_child = helpers.add_case(base_store, "case_child")
    case_child_id = case_child.internal_id
    helpers.add_relationship(
        store=base_store, case=case_child, sample=sample_child, mother=sample_mother
    )

    case_query = base_store._get_query(table=Case)
    family_sample_query = base_store._get_query(table=CaseSample)
    sample_query = base_store._get_query(table=Sample)

    assert case_query.count() == 2
    assert family_sample_query.count() == 2
    assert sample_query.count() == 2

    # WHEN deleting the child case
    result = cli_runner.invoke(delete_case_command, [case_child_id, "--yes"], obj=base_context)

    # THEN the child sample is deletable
    assert result.exit_code == SUCCESS
    assert case_query.count() == 1
    assert family_sample_query.count() == 1
    assert sample_query.count() == 1


def test_delete_trio_case(cli_runner: CliRunner, base_context: CGConfig, helpers):
    """Test that the delete case will delete a trio"""
    # GIVEN a database with a trio case
    base_store: Store = base_context.status_db
    case_obj = helpers.add_case(base_store)
    case_id = case_obj.internal_id
    sample_mother = helpers.add_sample(base_store, "mother")
    sample_father = helpers.add_sample(base_store, "father")
    sample_child = helpers.add_sample(base_store, "child")
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample_mother)
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample_father)
    helpers.add_relationship(
        store=base_store,
        case=case_obj,
        sample=sample_child,
        mother=sample_mother,
        father=sample_father,
    )
    case_query = base_store._get_query(table=Case)
    family_sample_query = base_store._get_query(table=CaseSample)
    sample_query = base_store._get_query(table=Sample)

    assert case_query.count() == 1
    assert family_sample_query.count() == 3
    assert sample_query.count() == 3

    # WHEN deleting a case
    result = cli_runner.invoke(delete_case_command, [case_id, "--yes"], obj=base_context)

    # THEN the trio case should be gone
    assert result.exit_code == SUCCESS
    assert case_query.count() == 0
    assert family_sample_query.count() == 0
    assert sample_query.count() == 0



from sqlalchemy.orm import Query

from cg.store.filters.status_pool_filters import (
    filter_pools_by_customer_ids,
    filter_pools_by_invoice_id,
    filter_pools_by_name_enquiry,
    filter_pools_by_order_enquiry,
    filter_pools_do_invoice,
    filter_pools_is_delivered,
    filter_pools_is_not_delivered,
    filter_pools_is_not_received,
    filter_pools_is_received,
    filter_pools_without_invoice_id,
)
from cg.store.models import Pool
from cg.store.store import Store
from tests.store.conftest import StoreConstants


def test_filter_pools_is_delivered(
    store_with_a_pool_with_and_without_attributes: Store,
    name=StoreConstants.NAME_POOL_WITH_ATTRIBUTES.value,
):
    """Test that a pool is returned when there is a delivered pool."""

    # GIVEN a store with two pools of which one is delivered

    # WHEN getting delivered pools
    pools: Query = filter_pools_is_delivered(
        pools=store_with_a_pool_with_and_without_attributes._get_query(table=Pool)
    )

    # ASSERT that the query is a Query
    assert isinstance(pools, Query)

    # THEN pools should contain the test pool
    assert pools.all()

    # THEN only one pool should be returned
    assert len(pools.all()) == 1

    # THEN the pool should have the expected name
    assert pools.all()[0].name == name


def test_filter_pools_is_not_delivered(
    store_with_a_pool_with_and_without_attributes: Store,
    name=StoreConstants.NAME_POOL_WITHOUT_ATTRIBUTES.value,
):
    """Test that a pool is returned when there is a pool that is not delivered."""

    # GIVEN a store with two pools of which one is delivered

    # WHEN getting not delivered pools
    pools: Query = filter_pools_is_not_delivered(
        pools=store_with_a_pool_with_and_without_attributes._get_query(table=Pool)
    )

    # ASSERT that the query is a Query
    assert isinstance(pools, Query)

    # THEN pools should contain the test pool
    assert pools.all()

    # THEN only one pool should be returned
    assert len(pools.all()) == 1

    # THEN the pool should have the expected name
    assert pools.all()[0].name == name


def test_filter_pools_is_received(
    store_with_a_pool_with_and_without_attributes: Store,
    name=StoreConstants.NAME_POOL_WITH_ATTRIBUTES.value,
):
    """Test that a pool is returned when there is a received pool."""

    # GIVEN a store with two pools of which one is received

    # WHEN getting received pools
    pools: Query = filter_pools_is_received(
        pools=store_with_a_pool_with_and_without_attributes._get_query(table=Pool)
    )

    # ASSERT that the query is a Query
    assert isinstance(pools, Query)

    # THEN pools should contain the test pool
    assert pools.all()

    # THEN only one pool should be returned
    assert len(pools.all()) == 1

    # THEN the pool should have the expected name
    assert pools.all()[0].name == name


def test_filter_pools_is_not_received(
    store_with_a_pool_with_and_without_attributes: Store,
    name=StoreConstants.NAME_POOL_WITHOUT_ATTRIBUTES.value,
):
    """Test that a pool is returned when there is a pool that is not received."""

    # GIVEN a store with two pools of which one is received

    # WHEN getting received pools
    pools: Query = filter_pools_is_not_received(
        pools=store_with_a_pool_with_and_without_attributes._get_query(table=Pool)
    )

    # ASSERT that the query is a Query
    assert isinstance(pools, Query)

    # THEN pools should contain the test pool
    assert pools.all()

    # THEN only one pool should be returned
    assert len(pools.all()) == 1

    # THEN the pool should have the expected name
    assert pools.all()[0].name == name


def test_filter_pools_do_invoice(
    store_with_a_pool_with_and_without_attributes: Store,
    name=StoreConstants.NAME_POOL_WITH_ATTRIBUTES.value,
):
    """Test that a pool is returned when there is a pool that should be invoiced."""

    # GIVEN a pool marked for invoicing and one not marked for invoicing

    # WHEN getting pools marked for invoicing
    pools: Query = filter_pools_do_invoice(
        pools=store_with_a_pool_with_and_without_attributes._get_query(table=Pool)
    )

    # ASSERT that the query is a Query
    assert isinstance(pools, Query)

    # THEN pools should contain the test pool
    assert pools.all()

    # THEN only one pool should be returned
    assert len(pools.all()) == 1

    # THEN the pool should have the expected name
    assert pools.all()[0].name == name


def test_filter_pools_by_invoice_id(
    store_with_a_pool_with_and_without_attributes: Store,
    name=StoreConstants.NAME_POOL_WITH_ATTRIBUTES.value,
    invoice_id=StoreConstants.INVOICE_ID_POOL_WITH_ATTRIBUTES.value,
):
    """Test that a pool is returned when there is a pool with a specific invoice id."""

    # GIVEN a store with two pools of which one us marked for invoicing

    # WHEN getting pools with invoice_id
    pools: Query = filter_pools_by_invoice_id(
        pools=store_with_a_pool_with_and_without_attributes._get_query(table=Pool),
        invoice_id=invoice_id,
    )

    # ASSERT that the query is a Query
    assert isinstance(pools, Query)

    # THEN pools should contain the test pool
    assert pools.all()

    # THEN only one pool should be returned
    assert len(pools.all()) == 1

    # THEN the pool should have the expected name
    assert pools.all()[0].name == name


def test_filter_pools_without_invoice_id(
    store_with_a_pool_with_and_without_attributes: Store,
    name=StoreConstants.NAME_POOL_WITHOUT_ATTRIBUTES.value,
):
    """Test that a pool is returned when there is a pool without invoice id."""

    # GIVEN a store with two pools of which one has no invoice id

    # WHEN getting pools without invoice_id
    pools: Query = filter_pools_without_invoice_id(
        pools=store_with_a_pool_with_and_without_attributes._get_query(table=Pool)
    )

    # ASSERT that the query is a Query
    assert isinstance(pools, Query)

    # THEN pools should contain the test pool
    assert pools.all()

    # THEN only one pool should be returned
    assert len(pools.all()) == 1

    # THEN the pool should have the expected name
    assert pools.all()[0].name == name


def test_filter_pools_by_name_enquiry(
    store_with_a_pool_with_and_without_attributes: Store,
    name=StoreConstants.NAME_POOL_WITH_ATTRIBUTES.value,
):
    """Test that a pool is returned when there is a pool with a specific name enquiry."""

    # GIVEN a store with two pools of which one has a name enquiry

    # WHEN getting pools with name enquiry
    pools: Query = filter_pools_by_name_enquiry(
        pools=store_with_a_pool_with_and_without_attributes._get_query(table=Pool),
        name_enquiry=StoreConstants.NAME_POOL_WITH_ATTRIBUTES.value,
    )

    # THEN a Query is returned
    assert isinstance(pools, Query)

    # THEN pools should contain the test pool
    assert pools.all()

    # THEN only one pool should be returned
    assert len(pools.all()) == 1

    # THEN the pool should have the expected name
    assert pools.all()[0].name == name


def test_filter_pools_by_order_enquiry(
    store_with_a_pool_with_and_without_attributes: Store,
    name=StoreConstants.NAME_POOL_WITH_ATTRIBUTES.value,
):
    """Test that a pool is returned when there is a pool with a specific order enquiry."""

    # GIVEN a store with two pools of which one has an order enquiry

    # WHEN getting pools with order enquiry
    pools: Query = filter_pools_by_order_enquiry(
        pools=store_with_a_pool_with_and_without_attributes._get_query(table=Pool),
        order_enquiry=StoreConstants.ORDER_POOL_WITH_ATTRIBUTES.value,
    )

    # THEN a Query is returned
    assert isinstance(pools, Query)

    # THEN pools should contain the test pool
    assert pools.all()

    # THEN only one pool should be returned
    assert len(pools.all()) == 1

    # THEN the pool should have the expected name
    assert pools.all()[0].name == name


def test_filter_pools_by_customer_id(
    store_with_a_pool_with_and_without_attributes: Store,
):
    """Test that a pool is returned when there is a pool with a specific customer id."""

    # GIVEN a store with two pools of with the same customer

    # WHEN getting pools with customer id
    pools: Query = filter_pools_by_customer_ids(
        pools=store_with_a_pool_with_and_without_attributes._get_query(table=Pool),
        customer_ids=[1],
    )

    # THEN a Query is returned
    assert isinstance(pools, Query)

    # THEN pools should be returned
    assert pools.all()

    # THEN two pools should be returned
    assert len(pools.all()) == 2

    # THEN the pool with the customer id should be returned
    assert pools.all()[0].customer_id == 1



"""Test MIP sample info file"""

from pathlib import Path

from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile
from cg.models.mip.mip_sample_info import MipBaseSampleInfo


def test_instantiate_mip_sampleinfo(sample_info_dna_raw: dict):
    """Tests sample info against a pydantic MipBaseSampleInfo"""
    # GIVEN a dictionary with the basic config

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**sample_info_dna_raw)

    # THEN assert that it was successfully created
    assert isinstance(sample_info_object, MipBaseSampleInfo)


def test_mip_sampleinfo(case_qc_sample_info_path: Path):
    """Test to parse the content of a real qc_sample_info file"""
    # GIVEN the path to a file with sample_info metadata content
    raw_sample_info = ReadFile.get_content_from_file(
        file_format=FileFormat.YAML, file_path=case_qc_sample_info_path
    )

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**raw_sample_info)

    # THEN assert that it was successfully created
    assert isinstance(sample_info_object, MipBaseSampleInfo)


def test_mip_sampleinfo_case_id(sample_info_dna_raw: dict):
    """Test case_id validator"""
    # GIVEN a dictionary with the sample info data

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**sample_info_dna_raw)

    # THEN assert that case_id was set
    assert sample_info_object.case_id == "yellowhog"


def test_mip_sampleinfo_case_id_with_family_id(sample_info_dna_raw: dict):
    """Test case_id validator"""
    # GIVEN a MIP sample info file missing a case_id but with family_id
    sample_info_dna_raw.pop("case_id")

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**sample_info_dna_raw)

    # THEN assert that case_id was set
    assert sample_info_object.case_id == "a_family_id"


def test_mip_sampleinfo_genome_build(sample_info_dna_raw: dict):
    """Test genome_build validator"""
    # GIVEN a dictionary with the sample info data

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**sample_info_dna_raw)

    # THEN assert that genome build was set
    assert sample_info_object.genome_build == "grch37"


def test_mip_sampleinfo_is_finished(sample_info_dna_raw: dict):
    """Test is_finished validator"""
    # GIVEN a dictionary with the sample info data

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**sample_info_dna_raw)

    # THEN assert that is_finished was set
    assert sample_info_object.is_finished


def test_mip_sampleinfo_is_finished(sample_info_dna_raw: dict):
    """Test is_finished validator"""
    # GIVEN a dictionary with the sample info data when analysisrunstatus is not finished
    sample_info_dna_raw["analysisrunstatus"] = "not_finished"

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**sample_info_dna_raw)

    # THEN assert that is_finished was set
    assert not sample_info_object.is_finished


def test_mip_sampleinfo_rank_model_version(sample_info_dna_raw: dict):
    """Test rank_model_version validator"""
    # GIVEN a dictionary with the sample info data

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**sample_info_dna_raw)

    # THEN assert that rank_model_version was set
    assert sample_info_object.rank_model_version == "v1.0"


def test_mip_sampleinfo_sv_rank_model_version(sample_info_dna_raw: dict):
    """Test sv_rank_model_version validator"""
    # GIVEN a dictionary with the sample info data

    # WHEN instantiating a MipBaseSampleInfo object
    sample_info_object = MipBaseSampleInfo(**sample_info_dna_raw)

    # THEN assert that rank_model_version was set
    assert sample_info_object.sv_rank_model_version == "v1.2.0"
